import binascii
import datetime
import itertools
import logging
import types
from collections import Counter, namedtuple
from random import randint

import six
import vstruct
from vstruct.primitives import *

import idb
import idb.netnode
import idb.typeinf
import idb.typeinf_flags

logger = logging.getLogger(__name__)

_counter = Counter()


def name_generator(prefix="unknown"):
    def inner(index=randint(0, 0xFFFFFFFF)):
        _counter[index] += 1
        return "{}{}".format(prefix, _counter[index])

    return inner


def is_flag_set(flags, flag):
    return flags & flag == flag


def as_unix_timestamp(buf, wordsize=None):
    """
    parse unix timestamp bytes into a timestamp.
    """
    q = struct.unpack_from("<I", buf, 0x0)[0]
    return datetime.datetime.utcfromtimestamp(q)


def as_md5(buf, wordsize=None):
    """
    parse raw md5 bytes into a hex-formatted string.
    """
    return binascii.hexlify(buf).decode("ascii")


def as_sha256(buf, wordsize=None):
    """
    parse raw sha256 bytes into a hex-formatted string.
    """
    return binascii.hexlify(buf).decode("ascii")


def cast(buf, V, wordsize=None):
    """
    apply a vstruct class to a sequence of bytes.

    Args:
        buf (bytes): the bytes to parse.
        V (type[vstruct.VStruct]): the vstruct class.

    Returns:
        V: the parsed instance of V.

    Example::

        s = cast(buf, Stat)
        assert s.gid == 0x1000
    """
    v = V(wordsize=wordsize)
    v.vsParse(buf)
    return v


def as_cast(V):
    """
    create a partial function that casts buffers to the given vstruct.

    Args:
        V (type[vstruct.VStruct]): the vstruct class.

    Returns:
        callable[bytes]->V: the function that parses buffers into V instances.

    Example::

        S = as_cast(Stat)
        s = S(buf)
        assert s.gid == 0x1000
    """

    def inner(buf, wordsize=None):
        return cast(buf, V, wordsize=wordsize)

    setattr(inner, "V", V.__name__)
    return inner


def unpack_dd(buf, offset=0):
    """
    unpack up to 32-bits using the IDA-specific data packing format.

    Args:
      buf (bytes): the region to parse.
      offset (int): the offset into the region from which to unpack. default: 0.

    Returns:
      (int, int): the parsed dword, and the number of bytes consumed.

    Raises:
      KeyError: if the bounds of the region are exceeded.
    """
    if offset != 0:
        # this isn't particularly fast... but its more readable.
        buf = buf[offset:]

    header = six.indexbytes(buf, 0x0)
    if header & 0x80 == 0:
        return header, 1
    elif header & 0xC0 != 0xC0:
        return ((header & 0x7F) << 8) + six.indexbytes(buf, 0x1), 2
    else:
        if header & 0xE0 == 0xE0:
            hi = (six.indexbytes(buf, 0x1) << 8) + six.indexbytes(buf, 0x2)
            low = (six.indexbytes(buf, 0x3) << 8) + six.indexbytes(buf, 0x4)
            size = 5
        else:
            hi = ((header & 0x3F) << 8) + six.indexbytes(buf, 0x1)
            low = (six.indexbytes(buf, 0x2) << 8) + six.indexbytes(buf, 0x3)
            size = 4
        return (hi << 16) + low, size


def unpack_dw(buf, offset=0):
    """
    unpack word.
    """
    if offset != 0:
        buf = buf[offset:]

    header = six.indexbytes(buf, 0x0)
    if header & 0x80 == 0:
        return header, 1
    elif header & 0xC0 != 0xC0:
        return ((header << 8) + six.indexbytes(buf, 0x1)) & 0x7FFF, 2
    else:
        return (six.indexbytes(buf, 0x1) << 8) + six.indexbytes(buf, 0x2), 3


def unpack_dq(buf, offset=0):
    """
    unpack qword.
    """
    if offset != 0:
        buf = buf[offset:]

    dw1, d1 = unpack_dd(buf)
    dw2, d2 = unpack_dd(buf, offset=d1)
    return (dw2 << 32) + dw1, d1 + d2


def unpack_dds(buf):
    offset = 0
    while offset < len(buf):
        val, size = unpack_dd(buf, offset=offset)
        yield val
        offset += size


def unpack_dqs(buf):
    offset = 0
    while offset < len(buf):
        val, size = unpack_dq(buf, offset=offset)
        yield val
        offset += size


class Unpacker:
    def __init__(self, buf, wordsize, offset=0, should_log=False):
        self.offset = offset
        self.wordsize = wordsize
        self.buf = buf
        self.should_log = should_log

    def _do_unpack(self, unpack_fn):
        v, delta = unpack_fn(self.buf, offset=self.offset)
        if self.should_log:
            logger.debug("%s at %x: %x", unpack_fn.__name__, self.offset, v)
        self.offset += delta
        return v

    def dd(self):
        return self._do_unpack(unpack_dd)

    def dq(self):
        return self._do_unpack(unpack_dq)

    def dw(self):
        return self._do_unpack(unpack_dw)

    def addr(self):
        if self.wordsize == 4:
            return self._do_unpack(unpack_dd)
        elif self.wordsize == 8:
            return self._do_unpack(unpack_dq)
        else:
            raise RuntimeError("unexpected wordsize")

    def off(self):
        offset = self.addr()
        mask = (2 ** (self.wordsize * 8)) - 1
        if offset & (1 << ((self.wordsize * 8) - 1)):
            return offset | ~mask
        else:
            return offset


Field = namedtuple("Field", ["name", "tag", "index", "cast", "minver"])
# namedtuple default args.
# via: https://stackoverflow.com/a/18348004/87207
Field.__new__.__defaults__ = (None,) * len(Field._fields)


class IndexType:
    def __init__(self, name):
        self.name = name

    def str(self):
        return self.name.upper()


ALL = IndexType("all")
ADDRESSES = IndexType("addresses")
NUMBERS = IndexType("numbers")
NODES = IndexType("nodes")

VARIABLE_INDEXES = (ALL, ADDRESSES, NUMBERS, NODES)


class _Analysis(object):
    """
    this is basically a metaclass for analyzers of IDA Pro netnode namespaces (named nodeid).
    provide set of fields, and parse them from netnodes (nodeid, tag, and optional index)
     when accessed.
    """

    def __init__(self, db, nodeid, fields):
        self.idb = db
        self.nodeid = nodeid
        self.netnode = idb.netnode.Netnode(db, nodeid)
        self.fields = fields

        idb_version = idb.netnode.Netnode(db, "Root Node").altval(index=-1)

        # note that order of fields is important:
        #   fields with matching minvers override previously defined fields of the same name
        self._fields_by_name = {
            f.name: f
            for f in self.fields
            if (not f.minver) or (f.minver and idb_version >= f.minver)
        }

    def _is_address(self, index):
        """
        does the given index fall within a segment?
        """
        try:
            self.idb.id1.get_segment(index)
            return True
        except KeyError:
            return False

    def _is_node(self, index):
        """
        does the index look like a raw nodeid?
        """
        if self.idb.wordsize == 4:
            return index & 0xFF000000 == 0xFF000000
        elif self.idb.wordsize == 8:
            return index & 0xFF00000000000000 == 0xFF00000000000000
        else:
            raise RuntimeError("unexpected wordsize")

    def _is_number(self, index):
        """
        does the index look like not (address or node)?
        """
        if self._is_node(index):
            return False

        if index < 0x1000:
            return True

        if self._is_address(index):
            return False

        return True

    def __getattr__(self, key):
        """
        for the given field name, fetch the value from the appropriate netnode.
        if the field matches multiple indices, then return a mapping from index to value.

        Example::

            assert root.version == 695

        Example::

            assert 0x401000 in entrypoints.ordinals

        Example::

            assert entrypoints.ordinals[0] == 'DllMain'

        Args:
          key (str): the name of the field to fetch.

        Returns:
          any: if a parser was provided, then the parsed data.
            otherwise, the bytes associatd with the field.
            if the field matches multiple indices, then the result is mapping from index to value.

        Raises:
          KeyError: if the field does not exist.
        """
        if key not in self._fields_by_name:
            return super(_Analysis, self).__getattribute__(key)

        field = self._fields_by_name[key]
        if field.index in VARIABLE_INDEXES:

            if field.index == ADDRESSES:
                nfilter = self._is_address
            elif field.index == NUMBERS:
                nfilter = self._is_number
            elif field.index == NODES:
                nfilter = self._is_node
            elif field.index == ALL:
                nfilter = lambda x: True
            else:
                raise ValueError("unexpected index")

            # indexes are variable, so map them to the values
            ret = {}
            for sup in self.netnode.supentries(tag=field.tag):
                if not nfilter(sup.parsed_key.index):
                    continue

                if field.cast is None:
                    ret[sup.parsed_key.index] = bytes(sup.value)
                else:
                    v = field.cast(bytes(sup.value), wordsize=self.idb.wordsize)
                    ret[sup.parsed_key.index] = v
            return ret
        else:
            # normal field with an explicit index
            v = self.netnode.supval(field.index, tag=field.tag)
            if field.cast is None:
                return bytes(v)
            else:
                return field.cast(bytes(v), wordsize=self.idb.wordsize)

    def get_field_tag(self, name):
        """
        get the tag associated with the given field name.

        Example::

            assert root.get_field_tag('version') == 'A'

        Args:
          key (str): the name of the field to fetch.

        Returns:
          str: a single character string tag.
        """
        return self._fields_by_name[name].tag

    def get_field_index(self, name):
        """
        get the index associated with the given field name.
        Example::

            assert root.get_field_index('version') == root.db.uint(-1)

        Args:
          key (str): the name of the field to fetch.

        Returns:
          int or IndexType: the index, if its specified.
            otherwise, this will be an `IndexType` that indicates what indices are expected.
        """
        return self._fields_by_name[name].index


def Analysis(nodeid, fields):
    """
    build a partial constructor for _Analysis with the given nodeid and fields.

    Example::

        Root = Analysis('Root Node', [Field(...), ...])
        root = Root(some_idb)
        assert root.version == 695
    """

    def inner(db):
        return _Analysis(db, nodeid, fields)

    return inner


class Reader(idb.typeinf.TypeString):
    def __init__(self, buf, wordsize):
        idb.typeinf.TypeString.__init__(self, buf)
        self.word = self.u32 if wordsize == 4 else self.u64
        self.word_ = self.u32_ if wordsize == 4 else self.u64_

    def bytes(self, size):
        return self.read(size)

    def str(self, size, encoding="utf-8"):
        return self.read(size).decode(encoding)

    def u16(self, big=False):
        if big:
            return struct.unpack(">H", self.read(2))[0]
        return struct.unpack("<H", self.read(2))[0]

    def u32(self, big=False):
        if big:
            return struct.unpack(">L", self.read(4))[0]
        return struct.unpack("<L", self.read(4))[0]

    def u64(self, big=False):
        if big:
            return struct.unpack(">Q", self.read(8))[0]
        return struct.unpack("<Q", self.read(8))[0]

    def u8_(self):
        return self.u8()

    def u16_(self):
        val = self.u8()
        if val == 0xFF:
            val = self.u16(big=True)
        elif val & 0x80:
            val = ((val & 0x3F) << 8) | self.u8()
        return val

    def u32_(self):
        val = self.u8()
        if val == 0xFF:
            val = self.u32(big=True)
        elif val & 0xC0 == 0xC0:
            val = (
                ((val & 0x1F) << 24) | (self.u8() << 16) | (self.u8() << 8) | self.u8()
            )
        elif val & 0x80:
            val = ((val & 0x3F) << 8) | self.u8()
        return val

    def u64_(self):
        val = self.u32_()
        val = val | (self.u32_() << 32)
        return val


class IdaInfo(vstruct.VStruct):
    """
    did not use vstruct to parse, but for compatibility, just inherit VStruct, and parse in vsParse()
    """

    # ref: https://www.hex-rays.com/products/ida/support/sdkdoc/structidainfo.html

    def __init__(self, wordsize):
        vstruct.VStruct.__init__(self)
        if wordsize in (4, 8):
            self.wordsize = wordsize
        else:
            raise ValueError("unexpected wordsize")

        """
        v7.0:
        nodeid: ff000002 tag: S index: 0x41b994
        00000000: 69 64 61 00 BC 02 6D 65  74 61 70 63 00 00 00 00  ida...metapc....
        00000010: 00 00 00 00 00 00 A3 00  0B 02 00 00 14 00 00 00  ................
        00000020: 0B 00 00 00 00 00 00 00  F7 FF FF DF 03 00 00 00  ................
        00000030: 00 00 00 00 FF FF FF FF  01 00 00 00 95 16 90 68  ...............h
        00000040: 95 16 90 68 FF FF FF FF  FF FF FF FF 00 10 90 68  ...h...........h
        00000050: 30 E2 9D 68 00 10 90 68  30 E2 9D 68 00 10 90 68  0..h...h0..h...h
        00000060: 00 70 9E 68 10 00 00 00  00 00 00 FF 00 00 10 FF  .p.h............
        00000070: 00 00 00 00 00 02 01 0F  0F 00 40 40 00 00 00 00  ..........@@....
        00000080: 00 00 00 00 00 00 00 00  00 00 02 06 67 BE A3 0E  ............g...
        00000090: 07 00 40 06 00 07 00 18  28 00 50 00 54 03 00 00  ..@.....(.P.T...
        000000A0: 01 00 00 00 01 1B 0A 00  00 00 00 00 61 00 00 00  ............a...
        000000B0: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  ................
        000000C0: 07 00 00 00 00 01 33 04  01 04 00 02 04 08 08 00  ......3.........
        000000D0: 00 00 00 00 00 00 00 00                           ........

        v6.95:
        00000000: 49 44 41 B7 02 6D 65 74  61 70 63 00 00 23 00 0B  IDA..metapc..#..
        00000010: 00 00 00 00 00 00 00 00  00 00 00 00 00 FF FF FF  ................
        00000020: FF FF FF 95 16 90 68 95  16 90 68 00 10 90 68 30  ......h...h...h0
        00000030: E2 9D 68 00 10 90 68 30  E2 9D 68 00 10 90 68 00  ..h...h0..h...h.
        00000040: 70 9E 68 10 00 00 00 0A  00 00 18 00 01 00 00 02  p.h.............
        00000050: 01 01 00 01 02 01 01 00  00 00 00 00 0F 08 00 09  ................
        00000060: 06 00 01 01 1B 07 61 00  00 00 00 00 00 00 00 00  ......a.........
        00000070: 00 00 00 00 00 00 00 00  00 00 00 01 00 00 00 01  ................
        00000080: 01 01 FF FF FF FF 01 00  00 00 FF FF FF FF 67 BE  ..............g.
        00000090: A3 0E 07 00 40 06 07 00  00 00 00 00 00 00 FD BF  ....@...........
        000000A0: 0F 00 28 00 50 00 40 40  00 00 00 00 00 00 00 00  ..(.P.@@........
        000000B0: 00 00 00 00 00 00 02 01  33 04 01 04 00 02 04 08  ........3.......
        000000C0: 14 00 00 00 08 00 00 00  00 00 00 00 00 00 00 00  ................
        000000D0: 00 00 00 00 00 00 00 00  00 00 00 00 00 01 00 00  ................
        000000E0: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  ................
        000000F0: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  ................
        """

        self.tag = None
        self.zero = None
        self.version = None
        self.procname_size = None
        self.procname = None
        self.lflags = None
        self.demnames = None
        self.filetype = None
        self.fcoresize = None
        self.corestart = None
        self.ostype = None
        self.apptype = None
        self.start_sp = None
        self.af = None
        self.start_ip = None
        self.begin_ea = None
        self.min_ea = None
        self.max_ea = None
        self.omin_ea = None
        self.omax_ea = None
        self.lowoff = None
        self.highoff = None
        self.maxref = None
        self.ascii_break = None
        self.wide_high_byte_first = None
        self.indent = None
        self.comment = None
        self.xrefnum = None
        self.entab = None
        self.specsegs = None
        self.voids = None
        self.showauto = None
        self.auto = None
        self.border = None
        self.null = None
        self.genflags = None
        self.showpref = None
        self.prefseg = None
        self.asmtype = None
        self.xrefs = None
        self.binpref = None
        self.cmtflag = None
        self.nametype = None
        self.showbads = None
        self.prefflag = None
        self.packbase = None
        self.asciiflags = None
        self.listnames = None
        self.asciipref = None
        self.asciisernum = None
        self.asciizeroes = None
        self.tribyte_order = None
        self.mf = None
        self.org = None
        self.assume = None
        self.checkarg = None
        self.start_ss = None
        self.start_cs = None
        self.main = None
        self.short_dn = None
        self.long_dn = None
        self.datatypes = None
        self.strtype = None
        self.af2 = None
        self.namelen = None
        self.margin = None
        self.lenxref = None
        self.lprefix = None
        self.lprefixlen = None
        self.compiler = None
        self.model = None
        self.sizeof_int = None
        self.sizeof_bool = None
        self.sizeof_enum = None
        self.sizeof_algn = None
        self.sizeof_short = None
        self.sizeof_long = None
        self.sizeof_llong = None
        self.change_counter = None
        self.sizeof_ldbl = None
        self.abiname = None
        self.abibits = None
        self.refcmts = None
        self.database_change_count = None

    def vsParse(self, sbytes, offset=0, fast=False):
        reader = Reader(sbytes, wordsize=self.wordsize)
        reader.pos = offset
        u8 = reader.u8
        u16 = reader.u16
        u32 = reader.u32
        u64 = reader.u64
        word = reader.word

        self.tag = reader.str(3)  # 'IDA' below 7.0, 'ida' in 7.0
        if self.tag == "ida":
            self.zero = u8()
        elif self.tag != "IDA":
            raise NotImplementedError("raise unknown database tag: " + self.tag)

        self.version = u16()
        if self.version == 700 and self.tag == "IDA":
            self.procname_size = u8()
        # 8 bytes for < 7.0
        # 16 bytes for > 7.0
        # 6.95 database upgraded to v7.0b
        # we have a single byte that describes how long the procname is.
        if self.procname_size is not None:
            self.procname = reader.str(self.procname_size)
        elif self.version < 700:
            self.procname = reader.str(8)
        elif self.version >= 700:
            self.procname = reader.str(16)
        import re

        self.procname = re.sub(r"\x00.*", "", self.procname)

        if self.version < 700:
            self.lflags = u8()
            self.demnames = u8()
            self.filetype = u16()
            self.fcoresize = word()  # 0x11
            self.corestart = word()
            self.ostype = u16()  # 0x19
            self.apptype = u16()
            self.start_sp = word()  # 0x1d
            self.af = u16()  # 0x21 Analysis Kernel options1+2
            self.start_ip = word()  # 0x23
            self.begin_ea = word()
            self.min_ea = word()
            self.max_ea = word()
            self.omin_ea = word()
            self.omax_ea = word()
            self.lowoff = word()
            self.highoff = word()
            self.maxref = word()
            self.ascii_break = u8()  # 71
            self.wide_high_byte_first = u8()
            self.indent = u8()
            self.comment = u8()
            self.xrefnum = u8()
            self.entab = u8()
            self.specsegs = u8()
            self.voids = u8()
            reader.seek(1)
            self.showauto = u8()  # 80
            self.auto = u8()
            self.border = u8()
            self.null = u8()
            self.genflags = u8()
            self.showpref = u8()
            self.prefseg = u8()
            self.asmtype = u8()
            self.baseaddr = word()  # 88
            self.xrefs = u8()
            self.binpref = u16()
            self.cmtflag = u8()  # 95
            self.nametype = u8()
            self.showbads = u8()
            self.prefflag = u8()
            self.packbase = u8()
            self.asciiflags = u8()
            self.listnames = u8()
            self.asciipref = reader.bytes(16)  # 102
            self.asciisernum = word()  # 118
            self.asciizeroes = u8()  # 122
            reader.seek(2)  # 123
            self.tribyte_order = u8()  # 125
            self.mf = u8()
            self.org = u8()
            self.assume = u8()
            self.checkarg = u8()
            self.start_ss = word()  # 131
            self.start_cs = word()
            self.main = word()
            self.short_dn = word()  # 143
            self.long_dn = word()
            self.datatypes = word()
            self.strtype = word()
            self.af2 = u16()  # 159
            self.namelen = u16()
            self.margin = u16()
            self.lenxref = u16()
            self.lprefix = reader.str(16)  # 167
            self.lprefixlen = u8()  # 183
            self.compiler = u8()
            self.model = u8()
            self.sizeof_int = u8()
            self.sizeof_bool = u8()
            self.sizeof_enum = u8()
            self.sizeof_algn = u8()
            self.sizeof_short = u8()
            self.sizeof_long = u8()
            self.sizeof_llong = u8()
            if len(sbytes) < 193:
                return reader.pos

            self.change_counter = u32()  # 193 0xc0
            self.sizeof_ldbl = u8()
            reader.seek(4)
            self.abiname = reader.str(size=16)
            self.abibits = u32()
            self.refcmts = u8()
        else:
            if self.tag == "IDA":
                u8 = reader.u8_
                u16 = reader.u16_
                u32 = reader.u32_
                u64 = reader.u64_
                word = reader.word_

            self.genflags = u16()
            self.lflags = u32()
            self.database_change_count = u32()
            self.filetype = u16()
            self.ostype = u16()
            self.apptype = u16()
            self.asmtype = u8()
            self.specsegs = u8()
            self.af = u32()
            self.af2 = u32()
            self.baseaddr = word()  # 48
            self.start_ss = word()
            self.start_cs = word()
            self.start_ip = word()
            self.start_ea = word()
            self.start_sp = word()
            self.main = word()
            self.min_ea = word()
            self.max_ea = word()
            self.omin_ea = word()
            self.omax_ea = word()
            self.lowoff = word()
            self.highoff = word()
            self.maxref = word()
            self.privrange_start_ea = word()
            self.privrange_end_ea = word()
            self.netdelta = word()
            self.xrefnum = u8()  # 116
            self.type_xrefnum = u8()
            self.refcmtnum = u8()
            self.xrefflag = u8()
            self.max_autoname_len = u16()  # 120

            if self.tag == "ida":
                reader.seek(17)

            self.nametype = u8()
            self.short_demnames = u32()  # 124
            self.long_demnames = u32()
            self.demnames = u8()  # 132
            self.listnames = u8()
            self.indent = u8()
            self.comment = u8()
            self.marzgin = u16()  # 136
            self.lenxref = u16()
            self.outflags = u32()  # 140
            self.cmtflg = u8()  # 144
            self.limiter = u8()
            self.bin_prefix_size = u16()  # 146
            self.prefflag = u8()  # 148
            self.strlit_flags = u8()
            self.strlit_break = u8()
            self.strlit_zeroes = u8()
            self.strtype = u32()

            self.strlit_pref_size = u8()
            if self.tag == "ida":
                self.strlit_pref = reader.str(16)
            else:
                self.strlit_pref = reader.str(self.strlit_pref_size)

            self.strlit_sernum = word()
            self.datatypes = word()
            self.cc_id = u8()
            self.cc_cm = u8()
            self.cc_size_i = u8()
            self.cc_size_b = u8()
            self.cc_size_e = u8()
            self.cc_defalign = u8()
            self.cc_size_s = u8()
            self.cc_size_l = u8()
            self.cc_size_ll = u8()
            self.cc_size_ldbl = u8()
            self.abibits = u32()
            self.appcall_options = u32()
        return reader.pos


Root = Analysis(
    "Root Node",
    [
        Field("imagebase", "A", -6, idb.netnode.as_int),
        Field("crc", "A", -5, idb.netnode.as_int),
        Field("open_count", "A", -4, idb.netnode.as_int),
        Field("created", "A", -2, as_unix_timestamp),
        Field("version", "A", -1, idb.netnode.as_int),
        Field("md5", "S", 1302, as_md5),
        Field("version_string", "S", 1303, idb.netnode.as_string),
        Field("sha256", "S", 1349, as_sha256),
        Field("idainfo", "S", 0x41B994, as_cast(IdaInfo)),
        Field("input_file_path", "V", None, idb.netnode.as_string),
    ],
)

Loader = Analysis(
    "$ loader name",
    [
        Field("plugin", "S", 0, idb.netnode.as_string),
        Field("format", "S", 1, idb.netnode.as_string),
    ],
)

# see `scripts/dump_user.py` for intepretation.
OriginalUser = Analysis("$ original user", [Field("data", "S", 0, bytes),])

# see `scripts/dump_user.py` for intepretation.
User = Analysis("$ user1", [Field("data", "S", 0, bytes),])


# this works for v6.95.
# for v7.0b, the data looks something like:
#
#     00000000: FF 68 90 10 00 C0 0D A0  00 90 00 00              .h..........
#
# which looks pack_dd/dq to me.
# TODO: need a way to detect versions and switch analysis implementations.


class FileRegion(vstruct.VStruct):
    def __init__(self, wordsize):
        vstruct.VStruct.__init__(self)
        if wordsize == 4:
            v_word = v_uint32
        elif wordsize == 8:
            v_word = v_uint64
        else:
            raise ValueError("unexpected wordsize")

        self.start = v_word()
        self.end = v_word()
        self.rva = v_uint32()


class FileRegionV70:
    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)
        self.start = u.addr()
        self.end = self.start + u.addr()
        self.rva = u.addr()


# '$ fileregions' maps from idb segment start address to details about it.
#
# supvals:
#   format1:
#     index: start effective address
#     value:
#       0x0: start effective address
#       0x4: end effective address
#       0x8: rva start?
FileRegions = Analysis(
    "$ fileregions",
    [
        Field("regions", "S", ADDRESSES, as_cast(FileRegion)),
        Field("regions", "S", ADDRESSES, FileRegionV70, minver=700),
    ],
)


class func_t:
    FUNC_TAIL = 0x00008000

    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)

        self.startEA = u.addr()
        self.endEA = self.startEA + u.addr()
        self.flags = u.dw()

        self.frame = None
        self.frsize = None
        self.frregs = None
        self.argsize = None
        self.owner = None
        self.refqty = None

        if not is_flag_set(self.flags, func_t.FUNC_TAIL):
            try:
                self.frame = u.addr()
                self.frsize = u.addr()
                self.frregs = u.dw()
                self.argsize = u.addr()
                # there is some other stuff here, based on IDB version/features
            except IndexError:
                # some of these we don't have, so we'll fall back to the default value of None.
                # eg. owner, refqty only present in some idb versions
                # eg. all of these, if high bit of flags not set.
                pass
        else:
            # We are in a function tail. Chunks can be above or below the tail
            # owner
            try:
                self.owner = self.startEA - u.off()
                self.refqty = u.dd()
            except IndexError:
                # see warning note above
                pass


# '$ funcs' maps from function effective address to details about it.
#
# supvals:
#   format1:
#     index: effective address
#     value: func_t
Functions = Analysis(
    "$ funcs",
    [
        Field("functions", "S", ADDRESSES, func_t),
        Field("comments", "C", ADDRESSES, idb.netnode.as_string),
        Field("repeatable_comments", "R", ADDRESSES, idb.netnode.as_string),
    ],
)


class PString(vstruct.VStruct):
    """
    short pascal string, prefixed with single byte length.
    """

    def __init__(self, length_is_total=True):
        vstruct.VStruct.__init__(self)
        self.length = v_uint8()
        self.s = v_str()
        self.length_is_total = length_is_total

    def pcb_length(self):
        length = self.length
        if self.length_is_total:
            length = length - 1
        self["s"].vsSetLength(length)


class TypeString(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.header = v_uint8()
        self.length = v_uint8()
        self.s = v_str()

    def pcb_header(self):
        if self.header != 0x3D:
            raise RuntimeError("unexpected type header")

    def pcb_length(self):
        length = self.length
        self["s"].vsSetLength(length - 1)


class StructMember:
    def __init__(self, db, identity):
        self.idb = db

        if isinstance(identity, six.integer_types):
            # if doesn't start with 0xFF0000..., add it.
            nodebase = idb.netnode.Netnode.get_nodebase(db)
            if identity < nodebase:
                identity += nodebase
            self.netnode = idb.netnode.Netnode(db, identity)
            self.nodeid = identity
        elif isinstance(identity, six.string_types):
            self.netnode = idb.netnode.Netnode(db, identity)
            self.nodeid = self.netnode.nodeid
        else:
            raise ValueError("Expected identify is integer or string")

    def get_fullname(self):
        return self.netnode.name()

    def get_name(self):
        return self.netnode.name().partition(".")[2]

    def get_typeinfo(self):
        return self.netnode.supval(tag="S", index=0x3000)

    def get_type(self):
        # nodeid: ff000078 tag: S index: 0x3000
        # 00000000: 3D 0A 48 49 4E 53 54 41  4E 43 45 00              =.HINSTANCE.

        try:
            v = self.netnode.supval(tag="S", index=0x3000)
            s = TypeString()
            s.vsParse(v)
            return s.s
        except KeyError:
            return None

    def get_enum_id(self):
        return self.netnode.altval(tag="A", index=0xB)

    def get_struct_id(self):
        return self.netnode.altval(tag="A", index=0x3)

    def get_member_comment(self):
        return self.netnode.supstr(tag="S", index=0x0)

    def get_repeatable_member_comment(self):
        return self.netnode.supstr(tag="S", index=0x1)

    # TODO: tag='A', index=0x10
    # TODO: tag='S', index=0x9, "ptrseg"

    def __str__(self):
        try:
            typ = self.get_type()
        except KeyError:
            return "StructMember(name: %s)" % (self.get_name())
        else:
            return "StructMember(name: %s, type: %s)" % (self.get_name(), typ)


class STRUCT_FLAGS:
    # via: https://www.hex-rays.com/products/ida/support/sdkdoc/group___s_f__.html

    # is variable size structure (varstruct)? More...
    SF_VAR = 0x00000001

    # is a union? More...
    SF_UNION = 0x00000002

    # has members of type "union"?
    SF_HASUNI = 0x00000004

    # don't include in the chooser list
    SF_NOLIST = 0x00000008

    # the structure comes from type library
    SF_TYPLIB = 0x00000010

    # the structure is collapsed
    SF_HIDDEN = 0x00000020

    # the structure is a function frame
    SF_FRAME = 0x00000040

    # alignment (shift amount: 0..31)
    SF_ALIGN = 0x00000F80

    # ghost copy of a local type
    SF_GHOST = 0x00001000


class Struct:
    """
    Example::

        struc = Struct(idb, 0xFF000075)
        assert struc.get_name() == 'EXCEPTION_INFO'
        assert len(struc.get_members()) == 5
        assert list(struc.get_members())[0].get_type() == 'DWORD'
    """

    def __init__(self, db, identity):
        """from https://github.com/nlitsme/pyidbutil/blob/7705bcde167fd34a5800bfe54ba99d195b44bbbb/idblib.py#L1380
            Decodes info for structures
            (structnode, N)          = structname
            (structnode, D, address) = xref-type
            (structnode, M, 0)       = packed struct info
            (structnode, S, 27)      = packed value(addr, byte)
        """

        self.idb = db

        if isinstance(identity, six.integer_types):
            # if doesn't start with 0xFF0000..., add it.
            nodebase = idb.netnode.Netnode.get_nodebase(db)
            if identity < nodebase:
                identity += nodebase
            self.netnode = idb.netnode.Netnode(db, identity)
            self.nodeid = identity
        elif isinstance(identity, six.string_types):
            self.netnode = idb.netnode.Netnode(db, identity)
            self.nodeid = self.netnode.nodeid
        else:
            raise ValueError("Expected identify is integer or string")

    def get_name(self):
        return self.netnode.name()

    def get_members(self):
        v = self.netnode.supval(tag="M", index=0)
        u = Unpacker(v, wordsize=self.idb.wordsize)
        flags = u.dd()
        count = u.dd()

        for i in range(count):
            nodeid_offset = u.addr()
            _ = u.addr()
            _ = u.addr()
            _ = u.dd()
            _ = u.dd()

            member_nodeid = self.netnode.nodebase + nodeid_offset
            yield StructMember(self.idb, member_nodeid)

    def find_member_by_name(self, name):
        for m in self.get_members():
            if m.get_name() == name:
                return m
        return None


def chunks(l, n):
    """
    Yield successive n-sized chunks from l.
    via: https://stackoverflow.com/a/312464/87207
    """
    if isinstance(l, types.GeneratorType):
        while True:
            v = list(itertools.islice(l, n))
            if not v:
                return
            yield v
    else:
        i = 0
        while True:
            try:
                v = l[i : i + n]
                yield v
            except IndexError:
                return
            i += n


def pairs(l):
    return chunks(l, 2)


Chunk = namedtuple("Chunk", ["effective_address", "length"])
FunctionParameter = namedtuple("FunctionParameter", ["type", "name"])
FunctionSignature = namedtuple(
    "FunctionSignature", ["calling_convention", "rtype", "unk", "parameters"]
)
StackChangePoint = namedtuple("StackChangePoint", ["effective_address", "change"])


def create_pstring_list(buf):
    _lst = []
    ofs = 0
    while ofs < len(buf.strip(b"\x00")):
        _len = struct.unpack_from("<B", buf, ofs)[0]
        _lst.append(buf[ofs : ofs + _len][1:].decode("utf-8"))
        ofs += _len
        if _len == 0:
            ofs += 1
    return _lst


class Function:
    """
    Example::

        func = Function(idb, 0x401000)
        assert func.get_name() == 'DllEntryPoint'
        assert func.get_signature() == '... DllEntryPoint(...)'
    """

    def __init__(self, db, fva):
        self.idb = db
        self.nodeid = fva
        self.netnode = idb.netnode.Netnode(db, self.nodeid)

    def get_name(self):
        try:
            return self.netnode.name()
        except KeyError:
            return "sub_%X" % (self.nodeid)

    def get_signature(self):
        try:
            typebuf = self.netnode.supval(tag="S", index=0x3000)
            namebuf = self.netnode.supval(tag="S", index=0x3001)
            typ = six.indexbytes(typebuf, 0x0)
            if not idb.typeinf_flags.is_type_func(typ):
                raise RuntimeError("is not function")

            names = create_pstring_list(namebuf)
            typedata = idb.typeinf.FuncTypeData()
            ts = idb.typeinf.TypeString(typebuf)
            typedata.deserialize(self.idb.til, ts, names, [])
            inf = Root(self.idb).idainfo

            return idb.typeinf.TInfo(
                typ, typedata, til=self.idb.til, name=self.get_name(), inf=inf
            )
        except KeyError:
            return None

    def get_chunks(self):
        v = self.netnode.supval(tag="S", index=0x7000)

        # stored as:
        #
        #   first chunk:
        #     effective addr
        #     length
        #   second chunk:
        #     delta from first.ea + first.length
        #     length
        #   third chunk:
        #     delta from second.ea + second.length
        #     length
        #   ...

        last_ea = 0
        last_length = 0

        if self.idb.wordsize == 4:
            unpacker = unpack_dds
        elif self.idb.wordsize == 8:
            unpacker = unpack_dqs
        else:
            raise RuntimeError("unexpected wordsize")

        for delta, length in pairs(unpacker(v)):
            ea = last_ea + last_length + delta
            yield Chunk(ea, length)
            last_ea = ea
            last_length = length

    # S-0x1000: sp change points
    # S-0x4000: register variables
    # S-0x5000: local labels
    # S-0x7000: function tails

    def get_stack_change_points(self):
        # ref: ida.wll@0x100793d0
        try:
            v = self.netnode.supval(tag="S", index=0x1000)
        except KeyError:
            return
        offset = self.nodeid

        if self.idb.wordsize == 4:
            unpacker = unpack_dds
        elif self.idb.wordsize == 8:
            unpacker = unpack_dqs
        else:
            raise RuntimeError("unexpected wordsize")
        for (delta, change) in pairs(unpacker(v)):
            offset += delta
            if change & 1:
                change = change >> 1
            else:
                change = -(change >> 1)

            yield StackChangePoint(offset, change)


Xref = namedtuple("Xref", ["frm", "to", "type"])


def _get_xrefs(db, tag, src=None, dst=None, types=None):
    if src is None and dst is None:
        raise ValueError("one of src or dst must be provided")

    nn = idb.netnode.Netnode(db, src if dst is None else dst)
    try:
        for entry in nn.charentries(tag=tag):
            if (types and entry.value in types) or (not types):
                if src is not None:
                    yield Xref(src, entry.parsed_key.index, entry.value)
                else:  # have dst
                    yield Xref(entry.parsed_key.index, dst, entry.value)
    except KeyError:
        return


def get_crefs_to(db, ea, types=None):
    """
    fetches the code references to the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    """
    return _get_xrefs(db, dst=ea, tag="X", types=types)


def get_crefs_from(db, ea, types=None):
    """
    fetches the code references from the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    """
    return _get_xrefs(db, src=ea, tag="x", types=types)


def get_drefs_to(db, ea, types=None):
    """
    fetches the data references to the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    """
    return _get_xrefs(db, dst=ea, tag="D", types=types)


def get_drefs_from(db, ea, types=None):
    """
    fetches the data references from the given address.

    Args:
      db (idb.IDB): the database.
      ea (int): the effective address from which to fetch xrefs.
      types (collection of int): if provided, a whitelist collection of xref types to include.

    Yields:
      int: xref address.
    """
    return _get_xrefs(db, src=ea, tag="d", types=types)


# under v6.95, this works.
class Fixup(vstruct.VStruct):
    def __init__(self, wordsize):
        vstruct.VStruct.__init__(self)
        # sizeof() == 0xB (fixed)
        # possible values: 0x0 - 0xC. top bit has some meaning.
        self.type = v_uint8()
        self.unk01 = v_uint16()  # this might be the segment index + 1?
        if wordsize == 4:
            self.offset = v_uint32()
            self.unk07 = v_uint32()
        elif wordsize == 8:
            self.unk03 = v_uint32()
            self.unk07 = v_uint16()
            self.offset = v_uint64()
        else:
            raise ValueError("unexpected wordsize")

    def pcb_type(self):
        if self.type != 0x04:
            raise NotImplementedError("fixup type %x not yet supported" % (self.type))

    def get_fixup_length(self):
        if self.type == 0x4:
            return 0x4
        else:
            raise NotImplementedError("fixup type %x not yet supported" % (self.type))


class FixupV70:
    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)

        # tbh, don't really know what these fields are...
        self.type = u.dw()
        self.unk1 = u.dd()
        self.unk2 = u.addr()
        self.offset = u.dd()  # strange this is not an offset

        if self.type != 0x8:
            raise NotImplementedError("fixup type %x not yet supported" % (self.type))

    def get_fixup_length(self):
        if self.type == 0x8:
            return 0x4
        else:
            raise NotImplementedError("fixup type %x not yet supported" % (self.type))


# '$ fixups' maps from fixup start address to details about it.
Fixups = Analysis(
    "$ fixups",
    [
        Field("fixups", "S", ADDRESSES, as_cast(Fixup)),
        Field("fixups", "S", ADDRESSES, FixupV70, minver=700),
    ],
)


def parse_seg_strings(buf, wordsize=None):
    strings = []
    offset = 0x0

    while offset < len(buf):
        if buf[offset] == 0x0:
            break

        string = PString(length_is_total=False)
        string.vsParse(buf, offset=offset)
        offset += len(string)
        strings.append(string.s)

    return strings


SegStrings = Analysis("$ segstrings", [Field("strings", "S", 0, parse_seg_strings),])


class Seg:
    def __init__(self, buf, wordsize):
        self.buf = buf
        u = Unpacker(buf, wordsize=wordsize)

        self.startEA = u.addr()
        self.endEA = self.startEA + u.addr()
        # index into `$ segstrings` array of strings.
        self.name_index = u.addr()

        # via: https://www.hex-rays.com/products/ida/support/sdkdoc/classsegment__t.html
        # use get/set_segm_class() functions
        self.sclass = u.addr()
        # this field is IDP dependent.
        # TODO: needs non-zero test
        self.orgbase = u.addr()
        # Segment flags
        self.flags = u.dd()
        # Segment alignment codes
        self.align = u.dd()
        # Segment combination codes
        self.comb = u.dd()
        # Segment permissions (0 means no information)
        self.perm = u.dd()
        # Number of bits in the segment addressing.
        # if 0: 16 bits
        # if 1: 32 bits
        # if 2: 64 bits
        # see: https://github.com/fireeye/flare-ida/blob/master/python/flare/jayutils.py#L94
        self.bitness = u.dd()
        # segment type (see Segment types). More...
        self.type = u.dd()
        # segment selector - should be unique.
        self.sel = u.dd()
        # default segment register values.
        self.defsr = NotImplementedError()
        # the segment color
        self.color = (u.dd() - 1) & 0xFFFFFFFF


# '$ segs' maps from segment start address to details about it.
#
# supvals:
#   format1:
#     index: start effective address
#     value: pack_dd data.
#         1: startEA
#         2: size
#         3: name index
#         ...
Segments = Analysis(
    "$ segs",
    [
        # we use `ALL` for the index type because `_is_address` above does not recognize
        #  addresses not backed by flags/bytes in the IDB.
        # there may be segments for the `.bss`, `extern`, etc sections here, and these
        #  do not have associated flags/bytes.
        # therefore, until we fix `is_address`, being slightly imprecise here works better.
        # note: all indexes in the `$ segs` netnode are addresses, so this assumption works ok.
        Field("segments", "S", ALL, Seg),
    ],
)

Imports = Analysis(
    "$ imports",
    [
        # index: entry number, value: node id
        Field("lib_netnodes", "A", NUMBERS, idb.netnode.as_uint),
        # index: entry number, value: dll name
        Field("lib_names", "S", NUMBERS, idb.netnode.as_string),
    ],
)

Import = namedtuple("Import", ["library", "function_name", "function_address"])


def enumerate_imports(db):
    """
    enumerate the functions imported by the module in the given database.

    yields:
      Tuple[str, str, int]: library name, function name, function address
    """
    imps = Imports(db)
    for index, libname in imps.lib_names.items():
        if index == 0xFFFFFFFF:
            continue

        # dereference the node id stored in the A val
        nnref = imps.lib_netnodes[index]
        nn = idb.netnode.Netnode(db, nnref)

        for funcaddr in nn.sups():
            try:
                funcname = nn.supstr(funcaddr)
                yield Import(libname, funcname, funcaddr)
            except KeyError:
                logger.warning("failed to find import supval: %x", funcaddr)
                continue


EntryPoints = Analysis(
    "$ entry points",
    [
        # index: ordinal, value: address, terminated by index: uint(-1)
        Field("functions", "A", NUMBERS, idb.netnode.as_uint),
        # index: address, value: address, should be only one?
        Field("main_entry", "A", ADDRESSES, idb.netnode.as_uint),
        # index: ordinal, value: ordinal
        Field("ordinals", "I", NUMBERS, idb.netnode.as_uint),
        # index: ordinal, value: string like (NTDLL!Rtl...)
        Field("forwarded_symbols", "F", NUMBERS, idb.netnode.as_string),
        # index: ordinal, value: string like (Rtl...)
        Field("function_names", "S", NUMBERS, idb.netnode.as_string),
        # index: address, value: string like (Rtl...), should be only one?
        Field("main_entry_name", "S", ADDRESSES, idb.netnode.as_string),
    ],
)

EntryPoint = namedtuple(
    "EntryPoint", ["name", "address", "ordinal", "forwarded_symbol"]
)


def enumerate_entrypoints(db):
    """
    enumerate the entry point functions in the given database.

    yields:
      Tuple[str, int, int, str]: function name, address, ordinal (optional), and forwarded symbol (optional)
    """
    ents = EntryPoints(db)

    ordinals = ents.ordinals
    forwarded_symbols = ents.forwarded_symbols
    names = ents.function_names
    names.update(ents.main_entry_name)

    for index, addr in ents.functions.items():
        if index == db.uint(-1):
            break
        yield EntryPoint(
            names.get(index), addr, ordinals.get(index), forwarded_symbols.get(index)
        )

    for index, addr in ents.main_entry.items():
        yield EntryPoint(
            names.get(index), addr, ordinals.get(index), forwarded_symbols.get(index)
        )


ScriptSnippets = Analysis(
    "$ scriptsnippets",
    [
        # number of spaces per tab
        Field("tabsize", "Y", 0x0, idb.netnode.as_uint),
        # netnode references
        Field("scripts", "A", NUMBERS, idb.netnode.as_uint),
    ],
)

ScriptSnippet = namedtuple("ScriptSnippet", ["name", "language", "code"])


def enumerate_script_snippets(db):
    """
    enumerate script snippets stored in the given database.

    yields:
      Tuple[str, str, str]: filename, language (Python or IDC), and source code
    """
    scripts = ScriptSnippets(db)

    for nnid in scripts.scripts.values():
        nn = idb.netnode.Netnode(db, nnid - 1)
        name = nn.supstr(0x0, tag="S")
        language = nn.supstr(0x1, tag="S")
        code = nn.supstr(0x0, tag="X")
        yield ScriptSnippet(name, language, code)



#!/usr/bin/env python3
"""
some documentation

author: Willi Ballenthin
email: willi.ballenthin@gmail.com
"""
import argparse
import logging
import sys

import hexdump

import idb
import idb.netnode

logger = logging.getLogger(__name__)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description="Dump an IDB B-tree to a textual representation."
    )
    parser.add_argument("idbpath", type=str, help="Path to input idb file")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable debug logging"
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Disable all output but errors"
    )
    args = parser.parse_args(args=argv)

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)
    elif args.quiet:
        logging.basicConfig(level=logging.ERROR)
        logging.getLogger().setLevel(logging.ERROR)
    else:
        logging.basicConfig(level=logging.INFO)
        logging.getLogger().setLevel(logging.INFO)

    with idb.from_file(args.idbpath) as db:
        cursor = db.id0.get_min()
        while True:
            if cursor.key[0] == 0x2E:
                try:
                    k = idb.netnode.parse_key(cursor.key, wordsize=db.wordsize)
                except UnicodeDecodeError:
                    hexdump.hexdump(cursor.key)
                else:
                    print(
                        "nodeid: %x tag: %s index: %s"
                        % (
                            k.nodeid,
                            k.tag,
                            hex(k.index) if k.index is not None else "None",
                        )
                    )
            else:
                hexdump.hexdump(cursor.key)

            hexdump.hexdump(bytes(cursor.value))
            print("--")

            try:
                cursor.next()
            except IndexError:
                break

    return 0


if __name__ == "__main__":
    sys.exit(main())



import contextlib

import six

# keep this here, so that its exposed as::
#
#    import idb.IDAPython
from idb.idapython import IDAPython

if six.PY2:

    def memview(buf):
        # on py2.7, we get this madness::
        #
        #     bytes(memoryview('foo')) == "<memoryview ...>"
        return buf


else:

    def memview(buf):
        return memoryview(buf)


@contextlib.contextmanager
def from_file(path):
    # break import cycle
    import idb.fileformat

    with open(path, "rb") as f:
        buf = memview(f.read())
        db = idb.fileformat.IDB(buf)
        db.vsParse(buf)
        yield db


def from_buffer(buf):
    # break import cycle
    import idb.fileformat

    buf = memview(buf)
    db = idb.fileformat.IDB(buf)
    db.vsParse(buf)
    return db



#!/usr/bin/env python
"""
Parse and display license information from an IDA Pro database.

author: Willi Ballenthin
email: willi.ballenthin@gmail.com
"""
import argparse
import datetime
import logging
import struct
import sys
import binascii

from vstruct.primitives import v_zstr_utf8

import idb
import idb.netnode

logger = logging.getLogger(__name__)


def is_encrypted(buf):
    return buf.find(b"\x00" * 4) >= 0x7F


HEXRAYS_PUBKEY = 0x93AF7A8E3A6EB93D1B4D1FB7EC29299D2BC8F3CE5F84BFE88E47DDBDD5550C3CE3D2B16A2E2FBD0FBD919E8038BB05752EC92DD1498CB283AA087A93184F1DD9DD5D5DF7857322DFCD70890F814B58448071BBABB0FC8A7868B62EB29CC2664C8FE61DFBC5DB0EE8BF6ECF0B65250514576C4384582211896E5478F95C42FDED


def decrypt(buf):
    """
    decrypt the given 1024-bit blob using Hex-Ray's public key.

    i'm not sure from where this public key originally came.
    the algorithm is derived from here:
        https://github.com/nlitsme/pyidbutil/blob/87cb3235a462774eedfafca00f67c3ce01eeb326/idbtool.py#L43

    Args:
      buf (bytes): at least 0x80 bytes, of which the first 1024 bits will be decrypted.

    Returns:
      bytes: 0x80 bytes of decrypted data.
    """
    enc = int(binascii.hexlify(buf[127::-1]), 16)
    dec = pow(enc, 0x13, HEXRAYS_PUBKEY)
    return binascii.a2b_hex("%0256x" % dec)


def parse_user_data(buf):
    """
    parse a decrypted user blob into a structured dictionary.

    Args:
      buf (bytes): exactly 0x80 bytes of plaintext data.

    Returns:
      Dict[str, Any]: a dictionary with the following values:
        - ts1 (datetime.datetime): timestamp in UTC of something. database creation?
        - ts2 (datetime.datetime): timestamp in UTC of something. sometimes zero.
        - id (str): the ID of the license.
        - name (str): the name of the user and organization that owns the license.
    """
    if len(buf) != 0x7F:
        raise ValueError("invalid user blob.")

    version = struct.unpack_from("<H", buf, 0x2)[0]
    if version == 0 or version > 750:
        raise NotImplementedError("user blob version not supported.")

    ts1, _, ts2 = struct.unpack_from("<III", buf, 0x10)
    id = "%02X-%02X%02X-%02X%02X-%02X" % struct.unpack_from("6B", buf, 0x1C)

    name = v_zstr_utf8()
    name.vsParse(buf[0x22:])

    return {
        # unknown if these are in UTC or not. right now, assuming so.
        "ts1": datetime.datetime.utcfromtimestamp(ts1),
        "ts2": datetime.datetime.utcfromtimestamp(ts2),
        "id": id,
        "name": name,
    }


def get_userdata(netnode):
    """
    fetch, decrypt, and parse the user data from the given netnode.

    Args:
      netnode (ida_netnode.Netnode): the netnode containing the user data.

    Returns:
      dict[str, Any]: see `parse_user_data`.
    """
    userdata = netnode.supval(0x0)

    if is_encrypted(userdata):
        userdata = decrypt(userdata)[1:]
    else:
        userdata = userdata[:0x7F]

    return parse_user_data(userdata)


def print_userdata(api, tag="$ original user"):
    try:
        netnode = api.ida_netnode.netnode(tag)
        data = get_userdata(netnode)
        print("user: %s" % data["name"])
        print("id:   %s" % data["id"])
        print("ts1:  %s" % data["ts1"].isoformat(" ") + "Z")
        print("ts2:  %s" % data["ts2"].isoformat(" ") + "Z")
    except KeyError:
        logger.warning("can' find {}".format(tag))
    except (NotImplementedError, ValueError) as e:
        logger.warning(e)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description="Parse and display license information from an IDA Pro database."
    )
    parser.add_argument("idbpath", type=str, help="Path to input idb file")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable debug logging"
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Disable all output but errors"
    )
    args = parser.parse_args(args=argv)

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)
    elif args.quiet:
        logging.basicConfig(level=logging.ERROR)
        logging.getLogger().setLevel(logging.ERROR)
    else:
        logging.basicConfig(level=logging.INFO)
        logging.getLogger().setLevel(logging.INFO)

    with idb.from_file(args.idbpath) as db:
        api = idb.IDAPython(db)
        print_userdata(api)
        print_userdata(api, "$ user1")
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
migrate from ida sdk and https://github.com/aerosoul94/tilutil
"""
import zlib
from abc import ABCMeta, abstractmethod

from cached_property import cached_property
from vstruct import VStruct
from vstruct.primitives import *

from idb.typeinf_flags import *


class TypeString:
    def __init__(self, buf, pos=0, parent=None):
        self.buf = buf
        self.pos = pos
        self.parent = parent

    def seek(self, n):
        self.pos += n
        if self.parent is not None:
            self.parent.seek(n)
        if self.pos > len(self.buf) or self.pos < 0:
            raise OverflowError("at offset {}".format(self.pos))

    def read(self, n):
        val = self.buf[self.pos : self.pos + n]
        self.seek(n)
        return val

    def unpack(self, unpack_fn, size=0):
        ret = unpack_fn(self.buf[self.pos :])
        if isinstance(ret, tuple):
            ret, _size = ret
            self.seek(_size if size == 0 else size)
        else:
            self.seek(size)
        return ret

    def peek_u8(self):
        return struct.unpack("<B", self.buf[self.pos : self.pos + 1])[0]

    def peek_u16(self):
        return struct.unpack("<H", self.buf[self.pos : self.pos + 2])[0]

    def u8(self):
        return struct.unpack("<B", self.read(1))[0]

    def u16(self):
        return struct.unpack("<H", self.read(2))[0]

    def db(self):
        """ read 1 byte as u8"""
        return self.u8()

    def dt(self):
        """ Reads 1 to 2 bytes.
        Value Range: 0-0xFFFE
        Usage: 16bit numbers
        :return: int
        """
        val = self.u8()
        if val & 0x80:
            val = val & 0x7F | self.u8() << 7
        return val - 1

    def de(self):
        """ Reads 1 to 5 bytes
        Value Range: 0-0xFFFFFFFF
        Usage: Enum Deltas
        :return: int
        """
        val = 0
        while True:
            hi = val << 6
            b = self.u8()
            sign = b & 0x80
            if not sign:
                lo = b & 0x3F
                val = lo | hi
                break
            else:
                lo = 2 * hi
                hi = b & 0x7F
                val = lo | hi
        return val

    def da(self):
        """ Reads 1 to 9 bytes.
        ValueRange: 0x-0x7FFFFFFF, 0-0xFFFFFFFF
        Usage: Arrays
        :return: (int, int)
        """
        a = 0
        b = 0
        da = 0
        base = 0
        nelem = 0
        while True:
            typ = self.peek_u8()
            if typ & 0x80 == 0:
                break
            self.seek(1)
            da = (da << 7) | typ & 0x7F
            b += 1
            if b >= 4:
                z = self.peek_u8()
                if z != 0:
                    base = 0x10 * da | z & 0xF
                nelem = (self.u8() >> 4) & 7
                while True:
                    y = self.peek_u8()
                    if (y & 0x80) == 0:
                        break
                    self.seek(1)
                    nelem = (nelem << 7) | y & 0x7F
                    a += 1
                    if a >= 4:
                        return True, nelem, base
        return False, nelem, base

    def pstring(self):
        length = self.dt()
        buf = self.read(length)
        return buf.decode("ascii")

    def pbytes(self):
        length = self.dt()
        buf = self.read(length)
        return buf

    def type_attr(self):
        val = 0
        tah = self.u8()
        tmp = ((tah & 1) | ((tah >> 3) & 6)) + 1
        if is_tah_byte(tah) or tmp == 8:
            if tmp == 8:
                val = tmp
            shift = 0
            while True:
                next_byte = self.u8()
                if next_byte == 0:
                    raise ValueError("type_attr(): failed to parse")
                val |= (next_byte & 0x7F) << shift
                if next_byte & 0x80 == 0:
                    break
                shift += 7
        unk = []
        if val & TAH_HASATTRS:
            val = self.dt()
            for _ in range(val):
                string = self.pstring()
                self.seek(self.dt())
                unk.append(string)
        return val

    def tah_attr(self):
        if self.has_next() and is_tah_byte(self.peek_u8()):
            return self.type_attr()
        return 0

    def sdacl_attr(self):
        if self.has_next() and is_sdacl_byte(self.peek_u8()):
            return self.type_attr()
        return 0

    def get(self):
        return TypeString(self.rest())

    def ref(self):
        return TypeString(self.rest(), parent=self)

    def rest(self):
        return self.buf[self.pos :]

    def has_next(self):
        return self.pos < len(self.buf)


def serialize_dt(n):
    if n > 0x7FFE:
        raise ValueError("Value too high for append_dt")
    lo = n + 1
    hi = n + 1
    result = bytearray()
    if lo > 127:
        result += struct.pack("<B", lo & 0x7F | 0x80)
        hi = (lo >> 7) & 0xFF
    result += struct.pack("<B", hi)
    return result


class TypeData:
    __metaclass__ = ABCMeta

    @abstractmethod
    def deserialize(self, til, type_string, fields, fieldcmts):
        pass


global_inf = None


class TInfo:
    def __init__(
        self, base_type=BT_UNK, type_details=None, til=None, name=None, inf=None
    ):
        self.base_type = base_type
        self.flags = 0
        self.type_details = type_details
        self.til = til
        self.name = name if name is not None else ""
        self.inf = inf
        self._types = til.types if til is not None else None

    def get_refname(self):
        if self.is_decl_typedef():
            nex = self.get_next_tinfo()
            if nex.base_type == 0:
                return (
                    "#{}".format(self.type_details.ordinal)
                    if self.type_details.is_ordref
                    else self.type_details.name
                )
            else:
                return nex.get_name()
        raise ValueError

    def get_name(self):
        if self.name == "" and self.is_decl_typedef():
            return self.get_refname()
        return self.name

    def get_next_tinfo(self):
        if self.is_decl_typedef() and self._types is not None:
            typedef_detail = self.type_details
            if typedef_detail.is_ordref:
                _def = self._types.get_by_ordinal(typedef_detail.ordinal)
            else:
                _def = self._types.find_by_name(typedef_detail.name)
            if _def is not None:
                return _def.type
            else:
                return TInfo()
        return TInfo()

    def get_final_tinfo(self):
        if self.is_decl_typedef() and self._types is not None:
            _type = self.get_next_tinfo()
            while _type.is_decl_typedef():
                _type = _type.get_next_tinfo()
            return _type
        return self

    def get_arr_object(self):
        if self.is_decl_array():
            return self.type_details.elem_type
        return TInfo()

    def get_pointed_object(self):
        if self.is_decl_ptr():
            pt = self.type_details
            if pt.closure is not None:
                return pt.closure
            else:
                return pt.obj_type
        return TInfo()

    def get_ptrarr_object(self):
        if self.is_decl_ptr_or_array():
            if self.is_decl_array():
                return self.get_arr_object()
            else:
                return self.get_pointed_object()
        return TInfo()

    def get_cc(self):
        if self.is_decl_func():
            return get_cc(self.type_details.cc)
        elif self.is_funcptr():
            return self.get_pointed_object().type_details.cc

    def get_rettype(self):
        if self.is_decl_func():
            return self.type_details.rettype
        elif self.is_funcptr():
            return self.get_pointed_object().type_details.rettype

    def get_size(self):
        # TODO:
        raise NotImplementedError()

    def get_conv(self):
        if not self.is_func():
            return ""
        cc = self.get_cc()
        conv = ""
        if cc == CM_CC_INVALID:
            conv += "__bad_cc"
        if cc == CM_CC_CDECL:
            conv = "__cdecl"
        # elif cc == CM_CC_ELLIPSIS:
        #     conv = "call"
        elif cc == CM_CC_STDCALL:
            conv = "__stdcall"
        elif cc == CM_CC_PASCAL:
            conv = "__pascal"
        elif cc == CM_CC_FASTCALL:
            conv = "__fastcall"
        elif cc == CM_CC_THISCALL:
            conv = "__thiscall"
        # elif cc == CM_CC_MANUAL:
        #     conv = "__manual"
        elif cc == CM_CC_SPOILED:
            conv = "__spoiled"
        elif cc in (CM_CC_SPECIALE, CM_CC_SPECIAL):
            conv += "__usercall"
        elif cc == CM_CC_SPECIALP:
            conv += "__userpurge"
        return conv

    def get_typename(self):
        global global_inf
        global_inf = self.inf if self.inf is not None else global_inf
        t = ""
        base = get_base_type(self.get_decltype())
        flags = get_type_flags(self.get_decltype())
        # 0-9
        if is_typeid_last(self.base_type):
            if base == BT_UNK:
                t += "unknown"
            elif base == BT_VOID:
                t += "void"
            # 2-7
            elif BT_INT8 <= base <= BT_INT:
                if self.is_unsigned():
                    t += "unsigned "
                if base == BT_INT8:
                    t += "int8"
                elif base == BT_INT16:
                    t += "int16"
                elif base == BT_INT32:
                    t += "int32"
                elif base == BT_INT64:
                    t += "int64"
                elif base == BT_INT128:
                    t += "int128"
                elif base == BT_INT:
                    t += "int"
            elif base == BT_BOOL:
                t += "bool"
            elif base == BT_FLOAT:
                if flags == BTMT_FLOAT:
                    t += "float"
                elif flags == BTMT_FLOAT:
                    t += "double"
                elif flags == BTMT_LNGDBL:
                    t += "long double"
                elif flags == BTMT_SPECFLT:
                    t += "special float"
                else:
                    t += "unknown float"
        else:
            if self.is_funcptr() or self.is_decl_func():
                if self.is_funcptr():
                    func = self.get_pointed_object()
                    star = "*"
                else:
                    func = self
                    star = ""

                t += func.get_rettype().get_typename() + " "

                conv = func.get_conv()
                t += "(" + conv
                if conv != "":
                    t += " "
                t += star + self.get_name()
                if func.is_user_cc() and not func.get_rettype().is_void():
                    loc = print_argloc(func.type_details.retloc)
                    if loc is not None:
                        t += "@<{}>".format(loc)
                t += ")"

                args = func.type_details.args
                t += "("
                for i in range(len(args)):
                    arg = args[i]
                    argtype = arg.type
                    t += argtype.get_typename()
                    t += " " + arg.name if arg.name != "" else ""
                    if self.is_user_cc():
                        loc = print_argloc(arg.argloc)
                        if loc is not None:
                            t += "@<{}>".format(loc)
                    if i < len(args) - 1:
                        t += ", "
                t += ")"
            elif self.is_decl_ptr():
                ptr = self.get_pointed_object()
                t += "{}*".format(ptr.get_typename())
            elif self.is_decl_array():
                n_elems = self.type_details.n_elems
                arrobj = self.get_arr_object()
                t += (
                    "{}[]".format(arrobj.get_typename())
                    if n_elems == 0
                    else "{}[{}]".format(self.get_arr_object().get_typename(), n_elems)
                )
            elif self.is_decl_udt() or self.is_decl_enum():
                ref = self.type_details.ref
                refname = "" if ref is None else ref.get_refname()
                t += self.get_name() if refname == "" else refname
            elif self.is_decl_bitfield():
                if self.type_details.is_unsigned:
                    t += "unsigned "
                t += "int{}".format(self.type_details.nbytes * 8)
            else:
                t += self.get_name()
        return t

    def get_typedeclare(self):
        t = ""
        if self.is_decl_typedef():
            t += "typedef {} {}".format(self.get_typename(), self.get_refname())
        elif self.is_decl_enum():
            t += "enum {}".format(self.get_name())
        elif self.is_decl_udt():
            if self.is_decl_union():
                t += "union "
            elif self.is_decl_struct():
                t += "struct "
            t += self.get_typename()
        else:
            t += self.get_typename()
        return t

    def get_typestr(self, indent=2):
        typestr = self.get_typedeclare()
        if self.is_decl_sue():
            members = self.type_details.members
            if self.is_decl_udt():
                if len(members) > 0 and members[0].is_baseclass():
                    typestr += " : {}".format(members[0].type.get_name())
                    members = members[1:]

                if len(members) == 0:
                    typestr += " { }"
                else:
                    typestr += "\n{\n"
                    for m in members:
                        typestr += " " * indent
                        typename = m.type.get_typename()
                        if m.type.is_funcptr():
                            typestr += (
                                typename.replace("{name}", m.name)
                                if m.name is not None
                                else typename
                            )
                            typestr += ";\n"
                        elif m.type.is_decl_bitfield():
                            typestr += "{} {} : {};\n".format(
                                typename, m.name, m.type.type_details.width
                            )
                        else:
                            typestr += "{} {};\n".format(typename, m.name)
                    typestr += "}"
            else:
                typestr += "\n{\n"
                for m in members:
                    typestr += " " * indent
                    typestr += "{} = 0x{:X},\n".format(m.name, m.value)
                typestr += "}"
        return typestr

    def has_details(self):
        return self.type_details is not None

    def has_vftable(self):
        raise NotImplementedError()

    def get_decltype(self):
        return self.base_type

    def get_realtype(self, full=True):
        if full:
            return self.get_final_tinfo().get_decltype()
        return self.get_decltype()

    def is_decl_typedef(self):
        return is_type_typedef(self.get_decltype())

    def is_decl_array(self):
        return is_type_array(self.get_decltype())

    def is_decl_bitfield(self):
        return is_type_bitfld(self.get_decltype())

    def is_decl_bool(self):
        return is_type_bool(self.get_decltype())

    def is_decl_char(self):
        return is_type_char(self.get_decltype())

    def is_decl_complex(self):
        return is_type_complex(self.get_decltype())

    def is_decl_const(self):
        return is_type_const(self.get_decltype())

    def is_decl_double(self):
        return is_type_double(self.get_decltype())

    def is_decl_enum(self):
        return is_type_enum(self.get_decltype())

    def is_decl_float(self):
        return is_type_float(self.get_decltype())

    def is_decl_floating(self):
        return is_type_floating(self.get_decltype())

    def is_decl_func(self):
        return is_type_func(self.get_decltype())

    def is_decl_int(self):
        return is_type_int(self.get_decltype())

    def is_decl_int128(self):
        return is_type_int128(self.get_decltype())

    def is_decl_int16(self):
        return is_type_int16(self.get_decltype())

    def is_decl_int32(self):
        return is_type_int32(self.get_decltype())

    def is_decl_int64(self):
        return is_type_int64(self.get_decltype())

    def is_decl_paf(self):
        return is_type_paf(self.get_decltype())

    def is_decl_partial(self):
        return is_type_partial(self.get_decltype())

    def is_decl_ptr(self):
        return is_type_ptr(self.get_decltype())

    def is_decl_ptr_or_array(self):
        return is_type_ptr_or_array(self.get_decltype())

    def is_decl_struct(self):
        return is_type_struct(self.get_decltype())

    def is_decl_sue(self):
        return is_type_sue(self.get_decltype())

    def is_decl_uchar(self):
        return is_type_uchar(self.get_decltype())

    def is_decl_udt(self):
        return is_type_struni(self.get_decltype())

    def is_decl_uint(self):
        return is_type_uint(self.get_decltype())

    def is_decl_uint128(self):
        return is_type_uint128(self.get_decltype())

    def is_decl_uint16(self):
        return is_type_uint16(self.get_decltype())

    def is_decl_uint32(self):
        return is_type_uint32(self.get_decltype())

    def is_decl_uint64(self):
        return is_type_uint64(self.get_decltype())

    def is_decl_union(self):
        return is_type_union(self.get_decltype())

    def is_decl_unknown(self):
        return is_type_unknown(self.get_decltype())

    def is_decl_void(self):
        return is_type_void(self.get_decltype())

    def is_decl_volatile(self):
        return is_type_volatile(self.get_decltype())

    # realtype

    def is_arithmetic(self):
        return is_type_arithmetic(self.get_realtype())

    def is_array(self):
        return is_type_array(self.get_realtype())

    def is_bitfield(self):
        return is_type_bitfld(self.get_realtype())

    def is_bool(self):
        return is_type_bool(self.get_realtype())

    def is_castable_to(self, target):
        raise NotImplementedError()

    def is_char(self):
        return is_type_char(self.get_realtype())

    def is_complex(self):
        return is_type_complex(self.get_realtype())

    def is_const(self):
        return is_type_const(self.get_realtype())

    def is_correct(self):
        raise NotImplementedError()

    def is_double(self):
        return is_type_double(self.get_realtype())

    def is_empty_udt(self):
        raise NotImplementedError()

    def is_enum(self):
        return is_type_enum(self.get_realtype())

    def is_ext_arithmetic(self):
        return is_type_ext_arithmetic(self.get_realtype())

    def is_ext_integral(self):
        return is_type_ext_integral(self.get_realtype())

    def is_float(self):
        return is_type_float(self.get_realtype())

    def is_floating(self):
        return is_type_floating(self.get_realtype())

    def is_forward_decl(self):
        raise NotImplementedError()

    def is_from_subtil(self):
        raise NotImplementedError()

    def is_func(self):
        return is_type_func(self.get_realtype())

    def is_funcptr(self):
        if not self.is_decl_ptr():
            return False
        typ = self.get_pointed_object()
        while typ.is_decl_ptr():
            typ = typ.get_pointed_object()
        return typ.is_decl_func()

    def is_high_func(self):
        raise NotImplementedError()

    def is_int(self):
        return is_type_int(self.get_realtype())

    def is_int128(self):
        return is_type_int128(self.get_realtype())

    def is_int16(self):
        return is_type_int16(self.get_realtype())

    def is_int32(self):
        return is_type_int32(self.get_realtype())

    def is_int64(self):
        return is_type_int64(self.get_realtype())

    def is_integral(self):
        return is_type_integral(self.get_realtype())

    def is_ldouble(self):
        return is_type_ldouble(self.get_realtype())

    def is_manually_castable_to(self, target):
        raise NotImplementedError()

    def is_one_fpval(self):
        raise NotImplementedError()

    def is_paf(self):
        return is_type_paf(self.get_realtype())

    def is_partial(self):
        return is_type_partial(self.get_realtype())

    def is_ptr(self):
        return is_type_ptr(self.get_realtype())

    def is_ptr_or_array(self):
        return is_type_ptr_or_array(self.get_realtype())

    def is_purging_cc(self):
        raise NotImplementedError()

    def is_pvoid(self):
        raise NotImplementedError()

    def is_scalar(self):
        raise NotImplementedError()

    def is_shifted_ptr(self):
        raise NotImplementedError()

    def is_signed(self):
        return get_type_flags(self.get_realtype()) == BTMT_SIGNED

    def is_small_udt(self):
        raise NotImplementedError()

    def is_sse_type(self):
        raise NotImplementedError()

    def is_struct(self):
        return is_type_struct(self.get_realtype())

    def is_sue(self):
        return is_type_sue(self.get_realtype())

    def is_uchar(self):
        return is_type_uchar(self.get_realtype())

    def is_udt(self):
        return is_type_struni(self.get_realtype())

    def is_uint(self):
        return is_type_uint(self.get_realtype())

    def is_uint128(self):
        return is_type_uint128(self.get_realtype())

    def is_uint16(self):
        return is_type_uint16(self.get_realtype())

    def is_uint32(self):
        return is_type_uint32(self.get_realtype())

    def is_uint64(self):
        return is_type_uint64(self.get_realtype())

    def is_union(self):
        return is_type_union(self.get_realtype())

    def is_unknown(self):
        return is_type_unknown(self.get_realtype())

    def is_unsigned(self):
        return get_type_flags(self.get_realtype()) == BTMT_UNSIGNED

    def is_user_cc(self):
        return is_user_cc(self.get_cc())

    def is_vararg_cc(self):
        raise NotImplementedError()

    def is_varstruct(self):
        raise NotImplementedError()

    def is_vftable(self):
        raise NotImplementedError()

    def is_void(self):
        return is_type_void(self.get_realtype())

    def is_volatile(self):
        return is_type_volatile(self.get_realtype())


class ErrorTInfo:
    def __init__(self, name):
        self.name = name

    def get_typestr(self, indent=2):
        return "{} Error".format(self.name)


def create_tinfo(til, type_info, fields=None, fieldcmts=None, name=None, inf=None):
    type_string = (
        type_info if isinstance(type_info, TypeString) else TypeString(type_info)
    )
    typ = type_string.peek_u8()
    tinfo = TInfo(typ, til=til, name=name, inf=inf)
    if is_typeid_last(typ) or get_base_type(typ) == BT_RESERVED:
        type_string.seek(1)
    else:
        type_data = None
        if is_type_ptr(typ):
            type_data = PointerTypeData()
        elif is_type_func(typ):
            type_data = FuncTypeData()
        elif is_type_array(typ):
            type_data = ArrayTypeData()
        elif is_type_typedef(typ):
            type_data = TypedefTypeData()
        elif is_type_struni(typ):
            type_data = UdtTypeData()
        elif is_type_enum(typ):
            type_data = EnumTypeData()
        elif is_type_bitfld(typ):
            type_data = BitfieldTypeData()
        type_data.deserialize(til, type_string, fields, fieldcmts)
        tinfo.type_details = type_data
    return tinfo


def create_ref(til, type_info, inf=None):
    if not type_info.startswith(b"="):
        type_info = b"=" + serialize_dt(len(type_info)) + type_info
    return create_tinfo(til, type_info, inf=inf)


class PointerTypeData(TypeData):
    """Representation of ptr_type_data_t"""

    def __init__(self):
        TypeData.__init__(self)
        self.obj_type = None
        self.closure = None
        self.based_ptr_size = 0
        self.taptr_bits = 0

    def deserialize(self, til, ts, fields, fieldcmts):
        typ = ts.u8()
        if is_type_closure(typ):
            if ts.u8() == RESERVED_BYTE:
                self.closure = create_tinfo(til, ts.ref())
            else:
                self.based_ptr_size = ts.u8()
        self.taptr_bits = ts.tah_attr()
        self.obj_type = create_tinfo(til, ts.ref(), fields, fieldcmts)
        return self


class ArrayTypeData(TypeData):
    def __init__(self):
        TypeData.__init__(self)
        self.elem_type = None  # tinfo_t
        self.base = None
        self.n_elems = 0

    def deserialize(self, til, ts, fields, fieldcmts):
        typ = ts.u8()
        if get_type_flags(typ) & BTMT_NONBASED:
            self.base = 0
            self.n_elems = ts.dt()
        else:
            _, self.n_elems, self.base = ts.da()
        ts.tah_attr()
        self.elem_type = create_tinfo(til, ts.ref(), fields, fieldcmts)
        return self


class FuncArg:
    def __init__(self):
        self.argloc = None  # argloc_t
        self.name = ""
        self.cmt = ""
        self.type = None  # tinfo_t
        self.flags = 0


# Register-relative argument location
class RRel:
    def __init__(self):
        self.off = 0
        self.reg = 0


class ArgLoc:
    def __init__(self):
        self.type = 0
        # union
        # {
        #   sval_t sval;                // ::ALOC_STACK, ::ALOC_STATIC
        #   uint32 reginfo;             // ::ALOC_REG1, ::ALOC_REG2
        #   rrel_t *rrel;               // ::ALOC_RREL
        #   scattered_aloc_t *dist;     // ::ALOC_DIST
        #   void *custom;               // ::ALOC_CUSTOM
        #   biggest_t biggest;          // to facilitate manipulation of this union
        # };
        self.sval = 0
        self.reginfo = 0
        self.rrel = None
        self.dist = []  # maybe ArgPart[]
        self.custom = None
        self.biggest = None

    def get_reg1(self):
        return self.reginfo & 0xFFFF

    def get_reg2(self):
        return (self.reginfo >> 16) & 0xFFFF

    def get_reginfo(self):
        return self.reginfo

    def get_stkoff(self):
        return self.sval


class ArgPart(ArgLoc):
    def __init__(self):
        ArgLoc.__init__(self)
        self.off = 0  # ushort
        self.size = 0  # ushort


def print_reg(ind):
    global global_inf
    regs = REGS[global_inf.procname] if global_inf is not None else REGS_METAPC
    r = regs[ind] if ind < len(regs) else None
    return "R{}".format(ind) if r is None else r


def print_argloc(argloc):
    typ = argloc.type
    t = ""
    if typ in (ALOC_STATIC, ALOC_STACK):
        return None
    elif typ == ALOC_REG1:
        t += print_reg(argloc.get_reg1())
    elif typ == ALOC_REG2:
        t += "{}:{}".format(print_reg(argloc.get_reg2()), print_reg(argloc.get_reg1()))
    elif typ == ALOC_DIST:
        dst = []
        for d in argloc.dist:
            dst.append("{}:{}".format(d.off, print_argloc(d)))
        t += ", ".join(dst)
    else:
        raise NotImplementedError
    return t


class RegInfo:
    def __init__(self):
        self.reg = 0
        self.size = 0


class FuncTypeData(TypeData):
    def __init__(self):
        TypeData.__init__(self)
        self.args = []
        self.flags = 0
        self.rettype = None  # tinfo_t
        self.retloc = None  # argloc_t
        self.stkargs = None  # uval_t
        self.spoiled = []  # reginfovec_t
        self.cc = 0

    def deserialize(self, til, ts, fields, fieldcmts):
        typ = ts.u8()
        self.flags |= 4 * get_type_flags(typ)

        cm = ts.peek_u8()
        if is_cc_spoiled(cm):
            while is_cc_spoiled(cm):
                ts.seek(1)
                nspoiled = cm & ~CM_CC_MASK
                if nspoiled == 15:
                    f = 2 * (ts.u8() & 0x1F)
                else:
                    for n in range(nspoiled):
                        reginfo = RegInfo()
                        b = ts.read_db()
                        if bool(b & 0x80):
                            size = ts.u8()
                            reg = b & 0x7F
                        else:
                            size = (b >> 4) + 1
                            reg = (b & 0xF) - 1
                        reginfo.size = size
                        reginfo.reg = reg
                        self.spoiled.append(reginfo)
                    f = 1
                cm = ts.peek_u8()
                self.flags |= f

        self.cc = ts.u8()

        ts.tah_attr()
        self.rettype = create_tinfo(til, ts.ref(), fields, fieldcmts)
        if is_cm_cc_special_pe(self.cc) and not self.rettype.is_void():
            self.retloc = self.deserialize_argloc(ts.ref())

        # args
        if is_cm_cc_voidarg(self.cc):
            return self
        n = ts.dt()
        for i in range(n):
            arg = FuncArg()
            if ts.has_next() and ts.peek_u8() == FAH_BYTE:
                ts.seek(1)
                arg.flags = ts.de()
            if fields is not None and i < len(fields):
                arg.name = fields[i]
            if fieldcmts is not None and i < len(fieldcmts):
                arg.cmt = fieldcmts[i]

            arg.type = create_tinfo(til, ts.ref(), fields, fieldcmts)
            if is_cm_cc_special_pe(self.cc):
                arg.argloc = self.deserialize_argloc(ts.ref())
            self.args.append(arg)
        return self

    def deserialize_argloc(self, ts):
        argloc = ArgLoc()
        t = ts.u8()
        if t == 0xFF:
            typ = ts.dt()
            typ0 = typ & 0xF
            argloc.type = typ0
            if typ0 in (ALOC_STACK, ALOC_STATIC):
                argloc.sval = ts.de()  # sval
            elif typ0 == ALOC_DIST:
                # maybe N?
                N = (typ >> 5) & 0x7
                for _ in range(N):
                    argpart = ArgPart()
                    argpart.type = ALOC_REG1
                    argpart.reginfo = ts.dt()
                    argpart.off = ts.dt()
                    argpart.size = ts.dt()
                    argloc.dist.append(argpart)
            elif typ0 == ALOC_REG1:
                argloc.reginfo = ts.dt() | (ts.dt << 0x100)
            elif typ0 == ALOC_REG2:
                argloc.reginfo = ts.dt()
            elif typ0 == ALOC_RREL:
                rrel = RRel()
                rrel.reg = ts.dt()  # rrel_t->reg
                rrel.off = ts.de()  # rrel_t->off
                argloc.rrel = rrel
            elif typ0 == ALOC_CUSTOM:
                raise NotImplementedError
        else:
            b = (t & 0x7F) - 1
            if t <= 0x80:
                if b >= 0:
                    argloc.type = ALOC_REG1
                    argloc.reginfo = b
                else:
                    argloc.type = ALOC_STACK
            else:
                c = ts.u8() - 1
                if c != -1:
                    argloc.type = ALOC_REG2
                    argloc.reginfo = b | (c << 16)
        return argloc


# tattr_field Type attributes for udt fields
TAFLD_BASECLASS = 0x0020  # field: do not include but inherit from the current field
TAFLD_UNALIGNED = 0x0040  # field: unaligned field
TAFLD_VIRTBASE = 0x0080  # field: virtual base (not supported yet)


class UdtMember:
    def __init__(self):
        self.offset = 0
        self.size = 0
        self.name = ""  # qstring
        self.cmt = ""  # qstring
        self.type = None  # tinfo_t
        self.effalign = 0
        self.tafld_bits = 0
        self.fda = 0

    def is_unaligned(self):
        return bool(self.tafld_bits & TAFLD_UNALIGNED)

    def is_baseclass(self):
        return bool(self.tafld_bits & TAFLD_BASECLASS)

    def is_virtbase(self):
        return bool(self.tafld_bits & TAFLD_VIRTBASE)


# tattr_udt Type attributes for udts
TAUDT_UNALIGNED = 0x0040  # struct: unaligned struct
TAUDT_MSSTRUCT = 0x0020  # struct: gcc msstruct attribute
TAUDT_CPPOBJ = 0x0080  # struct: a c++ object, not simple pod type


class UdtTypeData(TypeData):
    """An object to represent struct or union types"""

    def __init__(self):
        TypeData.__init__(self)
        self.members = []
        # total structure size in bytes
        self.total_size = 0
        # unpadded structure size in bytes
        self.unpadded_size = 0
        # effective structure alignment (in bytes)
        self.effalign = 0
        # TA... and TAUDT... bits.
        self.taudt_bits = 0
        # declared structure alignment (shift amount+1). 0 - unspecified
        self.sda = 0
        # pragma pack() alignment (shift amount)
        self.pack = 0
        self.is_union = False
        self.ref = None

    def deserialize(self, til, ts, fields, fieldcmts):
        typ = ts.u8()
        self.is_union = is_type_union(typ)
        n = ts.dt()
        if n == 0:
            self.ref = create_ref(til, ts.pbytes())
            self.taudt_bits = ts.sdacl_attr()
        else:
            if n == 0x7FFE:
                n = ts.de()
            alpow = n & 7
            member_cnt = n >> 3
            if alpow == 0:
                # inf.cc.defalign
                self.effalign = 0
            else:
                self.effalign = 1 << (alpow - 1)
            self.taudt_bits = ts.sdacl_attr()
            # process cpp class inherit
            field_i = 0
            for i in range(member_cnt):
                member = UdtMember()
                member.type = create_tinfo(til, ts.ref(), fields, fieldcmts)
                attr = ts.sdacl_attr() if not self.is_union else 0
                member.tafld_bits = attr
                member.fda = attr
                if not member.is_baseclass():
                    if len(fields) > field_i:
                        member.name = fields[field_i]
                    if n < len(fieldcmts):
                        member.cmt = fieldcmts[field_i]
                    field_i += 1
                self.members.append(member)
        return self


TAENUM_64BIT = 0x0020  # enum: store 64-bit values


class EnumMember:
    def __init__(self, name, value=0, cmt=""):
        self.name = name  # qstring
        self.cmt = cmt  # qstring
        self.value = value


class EnumTypeData(TypeData):
    """Representation of enum_type_data_t"""

    def __init__(self):
        TypeData.__init__(self)
        self.group_sizes = []  # intvec_t
        # TAENUM_64BIT   0x0020
        #  	enum: store 64-bit values
        self.taenum_bits = 0
        self.bte = 0
        self.members = []
        self.ref = None

    def deserialize(self, til, ts, fields, fieldcmts):
        typ = ts.u8()
        n = ts.dt()
        if n == 0:
            self.ref = create_ref(til, ts.pbytes())
            self.taenum_bits = ts.sdacl_attr()
        else:
            if n == 0x7FFE:
                n = ts.de()
            self.taenum_bits = ts.tah_attr()
            self.bte = ts.u8()
            cur = 0
            hi = 0
            mask = self.calc_mask(til)
            for i in range(n):
                # TODO: subarrays
                # https://www.hex-rays.com/products/ida/support/sdkdoc/group__tf__enum.html#ga9ae7aa54dbc597ec17cbb17555306a02
                lo = ts.de()
                if self.taenum_bits & TAENUM_64BIT:
                    hi = ts.de()
                if self.bte & BTE_BITFIELD:
                    self.group_sizes.append(ts.dt())
                cur += lo | (hi << 32) & mask
                member = EnumMember(fields[i], value=cur)
                self.members.append(member)
        return self

    def calc_mask(self, til):
        emsize = self.bte & BTE_SIZE_MASK
        if emsize != 0:
            bytesize = 1 << (emsize - 1)
        elif til is not None:
            bytesize = til.size_e
        else:
            bytesize = 4
        bitsize = bytesize * 8
        if bitsize < 64:
            return (1 << bitsize) - 1
        return 0xFFFFFFFFFFFFFFFF


class TypedefTypeData(TypeData):
    """Representation of typedef_type_data_t"""

    def __init__(self):
        TypeData.__init__(self)
        self.til = None
        # union
        # {
        #    const char* name is_ordref=false: target type name. we do not own this pointer!
        #    uint32   ordinal is_ordref=true: type ordinal number
        # }
        self.name = None
        self.ordinal = 0
        self.is_ordref = False
        self.resolve = False

    def deserialize(self, til, type_string, fields, fieldcmts):
        self.til = til
        typ = type_string.u8()
        buf = type_string.pbytes()
        if buf.startswith(b"#"):
            self.is_ordref = True
            self.ordinal = TypeString(buf[1:]).de()
        else:
            self.name = buf.decode("ascii")
        return self


class BitfieldTypeData(TypeData):
    """Representation of bitfield_type_data_t"""

    def __init__(self):
        TypeData.__init__(self)
        self.nbytes = 0
        self.width = 0
        self.is_unsigned = False

    def deserialize(self, til, type_string, fields, fieldcmts):
        typ = type_string.u8()
        self.nbytes = 1 << (get_type_flags(typ) >> 4)
        dt = type_string.dt()
        self.width = dt >> 1
        self.is_unsigned = bool(dt & 1)
        type_string.tah_attr()
        return self


class v_zbytes(v_zstr):
    """
    A v_zbytes placeholder class which will automatically return
    up to a null terminator bytes dynamically.
    """

    def vsGetValue(self):
        return self._vs_value[: -self._vs_align_pad]


class TILTypeInfo(VStruct):
    def __init__(self, format):
        VStruct.__init__(self)
        self.format = format

        self.flags = v_uint32()
        self.name = v_zstr_utf8()
        self.ordinal = v_uint32()
        self.type_info = v_zbytes()
        self.cmt = v_zstr_utf8()
        self.fields_buf = v_zbytes()
        self.fieldcmts = v_zbytes()
        self.sclass = v_uint8()

    def pcb_flags(self):
        # Format below 0x12 does not have 64 bit ordinal's
        if self.format < 0x12:
            self.flags &= 0x7FFFFFFF
        if self.flags >> 31:
            self.vsSetField("ordinal", v_uint64())
        if self.flags not in (0x7FFFFFFF, 0xFFFFFFFF):
            raise Exception("unsupported format {}".format(self.flags))

    def deserialize(self, til, inf):
        try:
            _type = create_tinfo(
                til,
                self.type_info,
                self.fields,
                self.fieldcmts,
                name=self.name,
                inf=inf,
            )
        except OverflowError:
            _type = ErrorTInfo(self.name)
        object.__setattr__(self, "type", _type)

    @cached_property
    def fields(self):
        fields = []
        pos = 0
        while pos < len(self.fields_buf):
            length = struct.unpack("<B", self.fields_buf[pos : pos + 1])[0]
            fields.append(self.fields_buf[pos + 1 : pos + length].decode("ascii"))
            pos += length
        fields = list(filter(lambda x: x != "", fields))
        return fields


class TILBucket(VStruct):
    def __init__(self, flags, format):
        VStruct.__init__(self)

        self.flags = flags
        self.format = format
        self.defs = None

        self.ndefs = v_uint32()
        self.size = v_uint32()

        if self.flags & TIL_ZIP:
            self.csize = v_uint32()
        else:
            self.csize = None

        self.buf = v_bytes()

    def pcb_size(self):
        self["buf"].vsSetLength(self.size)

    def pcb_csize(self):
        if self.csize is not None:
            self["buf"].vsSetLength(self.csize)

    def pcb_buf(self):
        if self.csize is not None:
            buf = zlib.decompress(self.buf)
            self.vsSetField("buf", buf)
        else:
            buf = self.buf.tobytes() if isinstance(self.buf, memoryview) else self.buf

        defs = []
        offset = 0
        for _ in range(self.ndefs):
            _def = TILTypeInfo(self.format)
            offset = _def.vsParse(buf, offset=offset)
            defs.append(_def)
        self.defs = defs

    def find_by_name(self, name):
        if not self.defs:
            return None
        _def = list(filter(lambda x: x.name == name, self.defs))
        if len(_def) == 0:
            return None
        return _def[0]

    @cached_property
    def ordinal_defs(self):
        return {i.ordinal: i for i in self.defs}

    def get_by_ordinal(self, ordinal):
        if ordinal not in self.ordinal_defs:
            return None
        return self.ordinal_defs[ordinal]


TIL_ZIP = 0x0001  # pack buckets using zip
TIL_MAC = 0x0002  # til has macro table
TIL_ESI = 0x0004  # extended sizeof info (short, long, longlong)
TIL_UNI = 0x0008  # universal til for any compiler
TIL_ORD = 0x0010  # type ordinal numbers are present
TIL_ALI = 0x0020  # type aliases are present (this bit is used only on the disk)
TIL_MOD = 0x0040  # til has been modified, should be saved
TIL_STM = 0x0080  # til has extra streams
TIL_SLD = 0x0100  # sizeof(long double)


class TIL(VStruct):
    def __init__(self, buf=None, wordsize=4, inf=None):
        VStruct.__init__(self)
        self.wordsize = wordsize
        self.inf = inf

        self.signature = v_str(size=0x06)

        # https://github.com/aerosoul94/tilutil/blob/master/distil.py#L545

        self.format = v_uint32()
        self.flags = v_uint32()

        self.title_len = v_uint8()
        self.title = v_str()

        self.base_len = v_uint8()
        self.base = v_str()

        self.id = v_uint8()
        self.cm = v_uint8()
        self.size_i = v_uint8()
        self.size_b = v_uint8()
        self.size_e = v_uint8()
        self.def_align = v_uint8()

        # self.size_s  uint8
        # self.size_l  uint8
        # self.size_ll  uint8
        # self.size_ldbl  uint8
        # self.syms  TILBucket
        # self.type_ordinal_numbers  uint32
        # self.types  TILBucket
        # self.macros  TILBucket

    def pcb_flags(self):
        if self.flags & TIL_ESI:
            self.vsAddField("size_s", v_uint8())
            self.vsAddField("size_l", v_uint8())
            self.vsAddField("size_ll", v_uint8())

        if self.flags & TIL_SLD:
            self.vsAddField("size_ldbl", v_uint8())

        self.vsAddField("syms", TILBucket(self.flags, self.format))

        if self.flags & TIL_ORD:
            self.vsAddField("type_ordinal_numbers", v_uint32())

        self.vsAddField("types", TILBucket(self.flags, self.format))
        self.vsAddField("macros", TILBucket(self.flags, self.format))

    def pcb_title_len(self):
        self["title"].vsSetLength(self.title_len)

    def pcb_base_len(self):
        self["base"].vsSetLength(self.base_len)

    def vsParse(self, sbytes, offset=0, fast=False):
        sbytes = sbytes.tobytes() if isinstance(sbytes, memoryview) else sbytes
        result = VStruct.vsParse(self, sbytes, offset, fast)

        self.types.defs.sort(key=lambda x: x.ordinal)
        self.deserialize_bucket(self.syms)
        self.deserialize_bucket(self.types)
        return result

    def deserialize_bucket(self, bucket):
        for t in bucket.defs:
            t.deserialize(self, self.inf)

    def validate(self):
        if self.signature != "IDATIL":
            raise ValueError("bad signature")
        return True



"""
IDAPython script that generates a YARA rule to match against the
basic blocks of the current function. It masks out relocation bytes
and ignores jump instructions (given that we're already trying to
match compiler-specific bytes, this is of arguable benefit).

If python-yara is installed, the IDAPython script also validates that
the generated rule matches at least one segment in the current file.

author: Willi Ballenthin <william.ballenthin@fireeye.com>
"""

import logging
from collections import namedtuple

import ida_funcs
import idaapi
import idautils
import idc

logger = logging.getLogger(__name__)

BasicBlock = namedtuple("BasicBlock", ["va", "size"])

# each rule must have at least this many non-masked bytes
MIN_BB_BYTE_COUNT = 4


def get_basic_blocks(fva):
    """
    return sequence of `BasicBlock` instances for given function.
    """
    ret = []
    func = ida_funcs.get_func(fva)
    if func is None:
        return ret

    for bb in idaapi.FlowChart(func):
        ret.append(BasicBlock(va=bb.startEA, size=bb.endEA - bb.startEA))

    return ret


def get_function(va):
    """
    return va for first instruction in function that contains given va.
    """
    return ida_funcs.get_func(va).startEA


Rule = namedtuple("Rule", ["name", "bytes", "masked_bytes"])


def is_jump(va):
    """
    return True if the instruction at the given address appears to be a jump.
    """
    return idc.GetMnem(va).startswith("j")


def bord(b):
    if isinstance(b, int):
        return b
    else:
        return ord(b)


def get_basic_block_rule(bb):
    """
    create and format a YARA rule for a single basic block.
    mask relocation bytes into unknown bytes (like '??').
    do not include final instructions if they are jumps.
    """
    # fetch the instruction start addresses
    insns = []
    va = bb.va
    while va < bb.va + bb.size:
        insns.append(va)
        va = idc.NextHead(va)

    # drop the last instruction if its a jump
    if is_jump(insns[-1]):
        insns = insns[:-1]

    bytes = []
    # `masked_bytes` is the list of formatted bytes,
    #   not yet join'd for performance.
    masked_bytes = []
    for va in insns:
        size = idc.ItemSize(va)
        if idaapi.contains_fixups(va, size):
            # fetch the fixup locations within this one instruction.
            fixups = []
            fixupva = idaapi.get_next_fixup_ea(va)
            fixups.append(fixupva)
            # TODO: assume the fixup size is four bytes, probably bad.
            fixupva += 4

            while fixupva < va + size:
                fixupva = idaapi.get_next_fixup_ea(fixupva)
                fixups.append(fixupva)
                # TODO: assume the fixup size is four bytes, probably bad.
                fixupva += 4

            # assume each fixup is four bytes (TODO!),
            #  and compute the addresses of each component byte.
            fixup_byte_addrs = set([])
            for fixup in fixups:
                for i in range(fixup, fixup + 4):
                    fixup_byte_addrs.add(i)

            # fetch and format each byte of the instruction,
            #  possibly masking it into an unknown byte if its a fixup.
            for i, byte in enumerate(idc.GetManyBytes(va, size)):
                byte_addr = i + va
                if byte_addr in fixup_byte_addrs:
                    bytes.append(bord(byte))
                    masked_bytes.append("??")
                else:
                    bytes.append(bord(byte))
                    masked_bytes.append("%02X" % (bord(byte)))
        elif "call" in idc.GetMnem(va):
            for i, byte in enumerate(idc.GetManyBytes(va, size)):
                bytes.append(bord(byte))
                masked_bytes.append("??")
        else:
            for byte in idc.GetManyBytes(va, size):
                bytes.append(bord(byte))
                masked_bytes.append("%02X" % (bord(byte)))

    return Rule("$0x%x" % (bb.va), bytes, masked_bytes)


def format_rules(fva, rules):
    """
    given the address of a function, and the byte signatures for basic blocks in
     the function, format a complete YARA rule that matches all of the
     basic block signatures.
    """
    name = idc.GetFunctionName(fva)

    # some characters aren't valid for YARA rule names
    safe_name = name
    BAD_CHARS = "@ /\\!@#$%^&*()[]{};:'\",./<>?"
    for c in BAD_CHARS:
        safe_name = safe_name.replace(c, "")

    md5 = idautils.GetInputFileMD5()
    ret = []
    ret.append("rule a_%s_%s {" % (md5, safe_name))
    ret.append("  meta:")
    ret.append('    sample_md5 = "%s"' % (md5))
    ret.append('    function_address = "0x%x"' % (fva))
    ret.append('    function_name = "%s"' % (name))
    ret.append("  strings:")
    for rule in rules:
        formatted_rule = " ".join(rule.masked_bytes)
        ret.append("    %s = { %s }" % (rule.name, formatted_rule))
    ret.append("  condition:")
    ret.append("    all of them")
    ret.append("}")
    return "\n".join(ret)


def create_yara_rule_for_function(fva):
    """
    given the address of a function, generate and format a complete YARA rule
     that matches the basic blocks.
    """
    rules = []
    for bb in get_basic_blocks(fva):
        rule = get_basic_block_rule(bb)

        # ensure there at least MIN_BB_BYTE_COUNT
        #  non-masked bytes in the rule, or ignore it.
        # this will reduce the incidence of many very small matches.
        unmasked_count = len(list(filter(lambda b: b != "??", rule.masked_bytes)))
        if unmasked_count < MIN_BB_BYTE_COUNT:
            continue

        rules.append(rule)

    return format_rules(fva, rules)


def get_segment_buffer(segstart):
    """
    fetch the bytes of the section that starts at the given address.
    if the entire section cannot be accessed, try smaller regions until it works.
    """
    segend = idaapi.getseg(segstart).endEA
    buf = None
    segsize = segend - segstart
    while buf is None:
        buf = idc.GetManyBytes(segstart, segsize - 1)
        if buf is None:
            segsize -= 0x1000
    return buf


Segment = namedtuple("Segment", ["start", "size", "name", "buf"])


def get_segments():
    """
    fetch the segments in the current executable.
    """
    for segstart in idautils.Segments():
        segend = idaapi.getseg(segstart).endEA
        segsize = segend - segstart
        segname = str(idc.SegName(segstart)).rstrip("\x00")
        segbuf = get_segment_buffer(segstart)
        yield Segment(segstart, segend, segname, segbuf)


class TestDidntRunError(Exception):
    pass


def test_yara_rule(rule):
    """
    try to match the given rule against each segment in the current exectuable.
    raise TestDidntRunError if its not possible to import the YARA library.
    return True if there's at least one match, False otherwise.
    """
    try:
        import yara
    except ImportError:
        logger.warning("can't test rule: failed to import python-yara")
        raise TestDidntRunError("python-yara not available")

    r = yara.compile(source=rule)

    for segment in get_segments():
        matches = r.match(data=segment.buf)
        if len(matches) > 0:
            logger.info("generated rule matches section: {:s}".format(segment.name))
            return True
    return False


def main():
    va = idc.ScreenEA()
    fva = get_function(va)
    rule = create_yara_rule_for_function(fva)
    print(rule)

    if test_yara_rule(rule):
        print("success: validated the generated rule")
    else:
        print("error: failed to validate generated rule")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logging.getLogger().setLevel(logging.INFO)
    main()



#!/usr/bin/env python3
"""
Extract scripts embedded within IDA Pro databases.

author: Willi Ballenthin
email: willi.ballenthin@gmail.com
"""
import argparse
import logging
import sys

import idb
import idb.netnode

logger = logging.getLogger(__name__)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description="Extract scripts embedded within IDA Pro databases."
    )
    parser.add_argument("idbpath", type=str, help="Path to input idb file")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable debug logging"
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Disable all output but errors"
    )
    args = parser.parse_args(args=argv)

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)
    elif args.quiet:
        logging.basicConfig(level=logging.ERROR)
        logging.getLogger().setLevel(logging.ERROR)
    else:
        logging.basicConfig(level=logging.INFO)
        logging.getLogger().setLevel(logging.INFO)

    with idb.from_file(args.idbpath) as db:
        try:
            for script in idb.analysis.enumerate_script_snippets(db):
                logger.debug("script: %s", script.name)
                logger.debug("language: %s", script.language)
                logger.debug("code: \n%s", script.code)
                if script.language == "Python":
                    ext = ".py"
                elif script.language == "IDC":
                    ext = ".idc"
                else:
                    raise ValueError("unexpected script language: " + script.language)

                filename = script.name + ext
                logger.info(
                    "writing %s script %s to %s", script.language, script.name, filename
                )
                with open(filename, "wb") as f:
                    f.write(script.code.encode("utf-8"))
        except KeyError:
            logger.warning("not found script snippets")

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import idaapi
import idautils
import idc


def print_section_list():
    for s in idautils.Segments():
        seg = idaapi.getseg(s)
        print("%s" % idc.SegName(s))
        print(" - start address: 0x%x" % seg.startEA)
        print(" - sclass: 0x%x" % seg.sclass)
        print(" - orgbase: 0x%x" % seg.orgbase)
        print(" - flags: 0x%x" % seg.flags)
        print(" - align: 0x%x" % seg.align)
        print(" - comb: 0x%x" % seg.comb)
        print(" - perm: 0x%x" % seg.perm)
        print(" - bitness: 0x%x" % seg.bitness)
        print(" - sel: 0x%x" % seg.sel)
        # print(' - defsr: 0x%x' % seg.defsr)
        print(" - type: 0x%x" % seg.type)
        print(" - color: 0x%x" % seg.color)


print_section_list()



#!/usr/bin/env python
import sys

from setuptools import setup, find_packages

# For Testing:
#
# python3.4 setup.py register -r https://testpypi.python.org/pypi
# python3.4 setup.py bdist_wheel upload -r https://testpypi.python.org/pypi
# python3.4 -m pip install -i https://testpypi.python.org/pypi
#
# For Realz:
#
# python3.4 setup.py register
# python3.4 setup.py bdist_wheel upload
# python3.4 -m pip install

PY3_DEPS = ["six", "hexdump", "vivisect-vstruct-wb>=1.0.3", "cached-property"]

# python2.7 has no `functools.lru_cache`,
# so use a backported copy when necessary.
PY2_DEPS = PY3_DEPS + ["functools32"]

if sys.version_info[0] == 2:
    DEPS = PY2_DEPS
elif sys.version_info[0] == 3:
    DEPS = PY3_DEPS
else:
    raise RuntimeError("unexpected python major version")

setup(
    name="python-idb",
    version="0.6.0",
    description="Pure Python parser for IDA Pro databases (.idb files)",
    author="Willi Ballenthin",
    author_email="willi.ballenthin@gmail.com",
    url="https://github.com/williballenthin/python-idb",
    license="Apache License 2.0",
    install_requires=DEPS,
    extras_require={
        # install like `pip install python-idb[disassembly]`
        # note, capstone is annoying to install on windows and in virtualenvs.
        "disassembly": ["capstone"],
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*"]),
    entry_points={"console_scripts": []},
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2.7",
    ],
)
